# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from pyqrllib.pyqrllib import hstr2bin, sha2_256

from collections import namedtuple
from qbitcoin.core.misc import logger
from qbitcoin.core.State import State
from qbitcoin.core.TransactionMetadata import TransactionMetadata
from qbitcoin.generated import qbit_pb2


class MultiSigAddressState(object):
    def __init__(self, protobuf_block=None):
        self._data = protobuf_block
        if protobuf_block is None:
            self._data = qbit_pb2.MultiSigAddressState()

        counter_mapping = namedtuple("counter_mapping", ["get", "update"])

        self._counter_by_name = {
            b'p_tx_hash': counter_mapping(self.transaction_hash_count,
                                          self.update_transaction_hash_count),
            b'p_multi_sig_spend': counter_mapping(self.multi_sig_spend_count,
                                                  self.update_multi_sig_spend_count),
        }

    @property
    def pbdata(self):
        """
        Returns a protobuf object that contains persistable data representing this object
        :return: A protobuf MultiSigAddressState object
        :rtype: qbit_pb2.MultiSigAddressState
        """
        return self._data

    @property
    def address(self):
        return self._data.address

    @property
    def creation_tx_hash(self):
        return self._data.creation_tx_hash

    @property
    def balance(self):
        return self._data.balance

    @property
    def signatories(self):
        return self._data.signatories

    @property
    def weights(self):
        return self._data.weights

    @property
    def threshold(self):
        return self._data.threshold

    # @property
    def transaction_hash_count(self):
        return self._data.transaction_hash_count

    def multi_sig_spend_count(self):
        return self._data.multi_sig_spend_count

    @staticmethod
    def generate_multi_sig_address(creation_tx_hash: bytes) -> bytes:
        desc = bytes(hstr2bin('110000'))
        prev_hash = bytes(sha2_256(desc + creation_tx_hash))
        new_hash = bytes(sha2_256(desc + prev_hash))[-4:]
        return desc + prev_hash + new_hash

    @staticmethod
    def create(creation_tx_hash: bytes,
               balance: int,
               signatories: list,
               weights: list,
               threshold: int,
               transaction_hash_count: int):
        multi_sig_address_state = MultiSigAddressState()

        multi_sig_address_state._data.address = MultiSigAddressState.generate_multi_sig_address(creation_tx_hash)
        multi_sig_address_state._data.creation_tx_hash = creation_tx_hash
        multi_sig_address_state._data.balance = balance
        multi_sig_address_state._data.signatories.extend(signatories)
        multi_sig_address_state._data.weights.extend(weights)
        multi_sig_address_state._data.threshold = threshold
        multi_sig_address_state._data.transaction_hash_count = transaction_hash_count

        return multi_sig_address_state

    @staticmethod
    def create_by_address(address: bytes):
        multi_sig_address_state = MultiSigAddressState()
        multi_sig_address_state._data.address = address
        return multi_sig_address_state

    def update_transaction_hash_count(self, value=1, subtract=False):
        if subtract:
            self._data.transaction_hash_count -= value
        else:
            self._data.transaction_hash_count += value

    def update_multi_sig_spend_count(self, value=1, subtract=False):
        if subtract:
            self._data.multi_sig_spend_count -= value
        else:
            self._data.multi_sig_spend_count += value

    def get_counter_by_name(self, name: bytes):
        return self._counter_by_name[name].get()

    def update_counter_by_name(self, name, value=1, subtract=False):
        self._counter_by_name[name].update(value, subtract)

    def update_balance(self, state_container, value, subtract=False):
        if subtract:
            self._data.balance -= value
        else:
            self._data.balance += value

    @staticmethod
    def get_default(creation_tx_hash, signatories: list, weights: list, threshold: int):
        return MultiSigAddressState.create(creation_tx_hash=creation_tx_hash,
                                           balance=0,
                                           signatories=signatories,
                                           weights=weights,
                                           threshold=threshold,
                                           transaction_hash_count=0)

    @staticmethod
    def address_is_valid(address: bytes) -> bool:
        # Multi-sig addresses start with 0x11 prefix and should be 29 bytes total
        if len(address) != 29:
            return False
            
        if address[0:1] != b'\x11':
            return False

        return True

    def serialize(self):
        return self._data.SerializeToString()

    def get_weight_by_signatory(self, signatory_address) -> [int, bool]:
        for i in range(len(self.signatories)):
            if self.signatories[i] == signatory_address:
                return self.weights[i], True
        return 0, False

    @staticmethod
    def get_multi_sig_address_state(state: State,
                                    multi_sig_tx):
        try:
            data = state._db.get_raw(multi_sig_tx.addr_from)
            pbdata = qbit_pb2.MultiSigAddressState()
            pbdata.ParseFromString(bytes(data))
            address_state = MultiSigAddressState(pbdata)
            return address_state
        except KeyError:
            return None

    @staticmethod
    def get_multi_sig_address_state_by_address(db,
                                               multi_sig_address: bytes):
        try:
            data = db.get_raw(multi_sig_address)
            pbdata = qbit_pb2.MultiSigAddressState()
            pbdata.ParseFromString(bytes(data))
            address_state = MultiSigAddressState(pbdata)
            return address_state
        except KeyError:
            return None

    @staticmethod
    def get_multi_sig_address_by_shared_key(state: State,
                                            shared_key: bytes):
        tx, _ = TransactionMetadata.get_tx_metadata(state, shared_key)
        if tx is None:
            return None
        return tx.multi_sig_address

    @staticmethod
    def remove_multi_sig_address_state(state: State,
                                       multi_sig_address: bytes,
                                       batch=None):
        try:
            state._db.delete(multi_sig_address, batch)
        except Exception as e:
            logger.warning("Exception ", e)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from pyqrllib.pyqrllib import bin2hstr
from collections import namedtuple

from qbitcoin.core import config
from qbitcoin.core.State import State
from qbitcoin.generated import qbit_pb2


class OptimizedAddressState(object):
    def __init__(self, protobuf_block=None):
        self._data = protobuf_block
        if protobuf_block is None:
            self._data = qbit_pb2.OptimizedAddressState()

        counter_mapping = namedtuple("counter_mapping", ["get", "update"])

        self._counter_by_name = {
            b"p_tx_hash": counter_mapping(self.transaction_hash_count,
                                          self.update_transaction_hash_count),
            b"p_tokens": counter_mapping(self.tokens_count,
                                         self.update_tokens_count),
            b"p_slaves": counter_mapping(self.slaves_count,
                                         self.update_slaves_count),
            b"p_lattice_pk": counter_mapping(self.lattice_pk_count,
                                             self.update_lattice_pk_count),
            b"p_multisig_address": counter_mapping(self.multi_sig_address_count,
                                                   self.update_multi_sig_address_count),
            b"p_multi_sig_spend": counter_mapping(self.multi_sig_spend_count,
                                                  self.update_multi_sig_spend_count),
            b"p_inbox_message": counter_mapping(self.inbox_message_count,
                                                self.update_inbox_message_count),
            b"p_falcon_pks": counter_mapping(self.falcon_pk_count,
                                            self.update_falcon_pk_count),
        }

    @staticmethod
    def bin_to_qaddress(binAddress):
        return 'Q' + bin2hstr(binAddress)

    @property
    def pbdata(self):
        """
        Returns a protobuf object that contains persistable data representing this object
        :return: A protobuf OptimizedAddressState object
        :rtype: qbit_pb2.OptimizedAddressState
        """
        return self._data

    @property
    def address(self):
        return self._data.address

    @property
    def height(self):
        """
        Returns the height calculated from the address
        Height is derived from the second byte of the address left-shifted by 1 bit
        """
        return self.address[1] << 1

    @property
    def nonce(self):
        return self._data.nonce

    @property
    def balance(self):
        return self._data.balance

    @property
    def used_falcon_pk_count(self):
        return self._data.used_falcon_pk_count

    @used_falcon_pk_count.setter
    def used_falcon_pk_count(self, new_value: int):
        self._data.used_falcon_pk_count = new_value

    # @property
    def transaction_hash_count(self):
        return self._data.transaction_hash_count

    # @property
    def tokens_count(self):
        return self._data.tokens_count

    # @property
    def slaves_count(self):
        return self._data.slaves_count

    # @property
    def lattice_pk_count(self):
        return self._data.lattice_pk_count

    # @property
    def multi_sig_address_count(self):
        return self._data.multi_sig_address_count

    def multi_sig_spend_count(self):
        return self._data.multi_sig_spend_count

    def inbox_message_count(self):
        return self._data.inbox_message_count

    def falcon_pk_count(self):
        return self._data.falcon_pk_count

    def get_counter_by_name(self, name: bytes):
        return self._counter_by_name[name].get()

    def update_counter_by_name(self, name, value=1, subtract=False):
        self._counter_by_name[name].update(value, subtract)

    def update_balance(self, state_container, value, subtract=False):
        if subtract:
            self._data.balance -= value
        else:
            self._data.balance += value

    @staticmethod
    def create(address: bytes,
               nonce: int,
               balance: int,
               transaction_hash_count: int,
               tokens_count: int,
               slaves_count: int,
               lattice_pk_count: int,
               multi_sig_address_count: int,
               falcon_pk_count: int = 0):
        address_state = OptimizedAddressState()

        address_state._data.address = address
        address_state._data.nonce = nonce
        address_state._data.balance = balance

        address_state._data.transaction_hash_count = transaction_hash_count
        address_state._data.tokens_count = tokens_count
        address_state._data.slaves_count = slaves_count
        address_state._data.lattice_pk_count = lattice_pk_count
        address_state._data.multi_sig_address_count = multi_sig_address_count
        address_state._data.falcon_pk_count = falcon_pk_count

        return address_state

    def update_transaction_hash_count(self, value=1, subtract=False):
        if subtract:
            self._data.transaction_hash_count -= value
        else:
            self._data.transaction_hash_count += value

    def update_tokens_count(self, value=1, subtract=False):
        if subtract:
            self._data.tokens_count -= value
        else:
            self._data.tokens_count += value

    def update_slaves_count(self, value=1, subtract=False):
        if subtract:
            self._data.slaves_count -= value
        else:
            self._data.slaves_count += value

    def update_lattice_pk_count(self, value=1, subtract=False):
        if subtract:
            self._data.lattice_pk_count -= value
        else:
            self._data.lattice_pk_count += value

    def update_multi_sig_address_count(self, value=1, subtract=False):
        if subtract:
            self._data.multi_sig_address_count -= value
        else:
            self._data.multi_sig_address_count += value

    def update_multi_sig_spend_count(self, value=1, subtract=False):
        if subtract:
            self._data.multi_sig_spend_count -= value
        else:
            self._data.multi_sig_spend_count += value

    def update_inbox_message_count(self, value=1, subtract=False):
        if subtract:
            self._data.inbox_message_count -= value
        else:
            self._data.inbox_message_count += value
            
    def update_falcon_pk_count(self, value=1, subtract=False):
        if subtract:
            self._data.falcon_pk_count -= value
        else:
            self._data.falcon_pk_count += value

    def increase_nonce(self):
        self._data.nonce += 1

    def decrease_nonce(self):
        self._data.nonce -= 1

    def get_slave_permission(self, slave_pk) -> int:
        slave_pk_str = str(slave_pk)
        if slave_pk_str in self._data.slave_pks_access_type:
            return self._data.slave_pks_access_type[slave_pk_str]

        return -1

    @staticmethod
    def get_default(address):
        address_state = OptimizedAddressState.create(address=address,
                                                     nonce=config.dev.default_nonce,
                                                     balance=config.dev.default_account_balance,
                                                     transaction_hash_count=0,
                                                     tokens_count=0,
                                                     slaves_count=0,
                                                     lattice_pk_count=0,
                                                     multi_sig_address_count=0,
                                                     falcon_pk_count=0)

        if address == config.dev.coinbase_address:
            address_state._data.balance = int(config.dev.max_coin_supply * config.dev.quark_per_qbitcoin)
        return address_state

    @staticmethod
    def address_is_valid(address: bytes) -> bool:
        # Only validate Falcon-512 addresses (25 bytes starting with 0x01)
        if len(address) == 25 and address[0:1] == b'\x01':
            # This is a Falcon-512 generated address from our wallet creator
            # Validate basic format: should have valid checksum
            return True
        
        # Invalid address format - only Falcon addresses are supported
        return False

    def serialize(self):
        return self._data.SerializeToString()

    @staticmethod
    def put_optimized_addresses_state(state: State, addresses_state: dict, batch=None):
        """
        :param addresses_state:
        :param batch:
        :return:
        """
        for address in addresses_state:
            address_state = addresses_state[address]
            data = address_state.pbdata.SerializeToString()
            state._db.put_raw(address_state.address, data, batch)

    @staticmethod
    def get_optimized_address_state(state: State, address: bytes):
        try:
            data = state._db.get_raw(address)
            pbdata = qbit_pb2.OptimizedAddressState()
            pbdata.ParseFromString(bytes(data))
            address_state = OptimizedAddressState(pbdata)
            return address_state
        except KeyError:
            return OptimizedAddressState.get_default(address)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Qbitcoin Node GUI Application
A comprehensive desktop GUI for managing and monitoring a Qbitcoin blockchain node
using the Flet framework with full wallet integration, P2P settings, API management,
and real-time blockchain monitoring.
"""

import sys
import os
import json
import threading
import time
import datetime
import logging
import subprocess
from typing import Dict
import grpc

# Add qbitcoin path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '.')))

import flet as ft
from qbitcoin.core import config
from qbitcoin.tools.wallet_creator import WalletCreator
from qbitcoin.generated import qbit_pb2, qbit_pb2_grpc, qbitwallet_pb2, qbitwallet_pb2_grpc, qbitmining_pb2_grpc
from qbitcoin.core.misc import logger
from pyqrllib.pyqrllib import bin2hstr, hstr2bin

# Import modal dialog components
from modals import BlockDetailsModal, TransactionDetailsModal, WalletDetailsModal


class GUILogHandler(logging.Handler):
    """Custom log handler to capture logs and send them to GUI"""
    
    def __init__(self, gui_instance):
        super().__init__()
        self.gui_instance = gui_instance
        
    def emit(self, record):
        try:
            msg = self.format(record)
            level = record.levelname
            if self.gui_instance:
                self.gui_instance.add_log(msg, level)
        except Exception:
            self.handleError(record)


class QbitcoinGUI:
    """Main GUI Application for Qbitcoin Node Management"""
    
    def __init__(self, page: ft.Page):
        self.page = page
        self.page.title = "Qbitcoin Node Manager"
        self.page.window_width = 1400
        self.page.window_height = 900
        self.page.theme_mode = ft.ThemeMode.DARK  # Force dark theme
        
        # Node and services
        self.node_running = False
        self.wallet_daemon = None
        self.node_process = None  # For the main node process
        
        # gRPC stubs for API communication
        self.public_stub = None
        self.wallet_stub = None
        self.mining_stub = None
        
        # UI State
        self.blockchain_data = []
        self.peer_data = []
        self.transaction_pool = []
        self.debug_logs = []
        
        # Settings
        self.settings = self.load_settings()
        
        # Initialize modal dialogs
        self.block_details_modal = None
        self.transaction_details_modal = None
        self.wallet_details_modal = None
        
        # Initialize log viewer first so logs can be added during setup
        self.log_viewer = ft.ListView(
            height=400,
            spacing=2,
            auto_scroll=True,
            controls=[]
        )
        
        # Setup logging to capture node logs
        self.setup_logging()
        
        # Initialize UI
        self.setup_page()
        self.create_main_layout()
        
        # Initialize modal dialogs
        self.initialize_modals()
        
        # Start background monitoring
        self.start_monitoring()
    
    def load_settings(self) -> Dict:
        """Load application settings"""
        default_settings = {
            "mining_enabled": False,
            "mining_threads": 1,
            "mining_address": "",
            "p2p_port": 19000,
            "api_enabled": True,
            "api_port": 19009,
            "wallet_path": os.path.join(config.user.wallet_dir, "qbitcoin_wallet.json"),
            "node_data_dir": config.user.qrl_dir,
            "log_level": "INFO",
            "auto_start": False
        }
        
        settings_file = os.path.join(config.user.qrl_dir, "gui_settings.json")
        try:
            if os.path.exists(settings_file):
                with open(settings_file, 'r') as f:
                    loaded_settings = json.load(f)
                default_settings.update(loaded_settings)
        except Exception as e:
            print(f"Error loading settings: {e}")
        
        return default_settings
    
    def save_settings(self):
        """Save application settings"""
        settings_file = os.path.join(config.user.qrl_dir, "gui_settings.json")
        try:
            os.makedirs(os.path.dirname(settings_file), exist_ok=True)
            with open(settings_file, 'w') as f:
                json.dump(self.settings, f, indent=2)
        except Exception as e:
            print(f"Error saving settings: {e}")
    
    def connect_grpc_services(self):
        """Connect to gRPC services"""
        try:
            # Connect to public API
            public_channel = grpc.insecure_channel(f'{config.user.public_api_host}:{config.user.public_api_port}')
            self.public_stub = qbit_pb2_grpc.PublicAPIStub(public_channel)
            
            # Connect to wallet API (if wallet daemon is running)
            if self.wallet_daemon:
                wallet_channel = grpc.insecure_channel(f'{config.user.wallet_api_host}:{config.user.wallet_api_port}')
                self.wallet_stub = qbitwallet_pb2_grpc.WalletAPIStub(wallet_channel)
            
            # Connect to mining API
            mining_channel = grpc.insecure_channel(f'{config.user.mining_api_host}:{config.user.mining_api_port}')
            self.mining_stub = qbitmining_pb2_grpc.MiningAPIStub(mining_channel)
            
            self.add_log("Connected to gRPC services")
            return True
            
        except Exception as e:
            self.add_log(f"Error connecting to gRPC services: {e}", "ERROR")
            return False
    
    def get_node_status(self):
        """Get node status from API"""
        try:
            if self.public_stub:
                request = qbit_pb2.GetNodeStateReq()
                response = self.public_stub.GetNodeState(request, timeout=5)
                return response.info
            return None
        except Exception as e:
            self.add_log(f"Error getting node status: {e}", "ERROR")
            return None
    
    def get_wallet_info(self):
        """Get wallet information"""
        try:
            if self.wallet_stub:
                request = qbitwallet_pb2.GetWalletInfoReq()
                response = self.wallet_stub.GetWalletInfo(request, timeout=5)
                return response
            return None
        except Exception as e:
            self.add_log(f"Error getting wallet info: {e}", "ERROR")
            return None
    
    def get_addresses_list(self):
        """Get list of wallet addresses"""
        try:
            if self.wallet_stub:
                request = qbitwallet_pb2.ListAddressesReq()
                response = self.wallet_stub.ListAddresses(request, timeout=5)
                return response.addresses
            return []
        except Exception as e:
            self.add_log(f"Error getting addresses: {e}", "ERROR")
            return []
    
    def get_balance(self, address: str):
        """Get balance for an address"""
        try:
            if self.wallet_stub:
                request = qbitwallet_pb2.BalanceReq(address=address)
                response = self.wallet_stub.GetBalance(request, timeout=5)
                return response.balance
            return "0"
        except Exception as e:
            self.add_log(f"Error getting balance: {e}", "ERROR")
            return "0"
    
    def setup_page(self):
        """Setup page configuration"""
        self.page.bgcolor = ft.Colors.GREY_900  # Dark background
        self.page.scroll = ft.ScrollMode.AUTO
        
        # Status bar
        self.status_text = ft.Text("Ready", size=12, color=ft.Colors.GREEN)
        self.status_bar = ft.Container(
            content=ft.Row([
                ft.Icon(ft.Icons.CIRCLE, color=ft.Colors.RED, size=12),
                self.status_text,
                ft.Text(f"Qbitcoin Node GUI v{config.dev.version}", size=10, color=ft.Colors.GREY_400)
            ]),
            padding=5,
            bgcolor=ft.Colors.GREY_800,  # Dark status bar
            height=30
        )
    
    def create_main_layout(self):
        """Create the main application layout"""
        # Main tabs
        self.tabs = ft.Tabs(
            selected_index=0,
            animation_duration=300,
            tabs=[
                ft.Tab(text="Dashboard", icon=ft.Icons.DASHBOARD, content=self.create_dashboard_tab()),
                ft.Tab(text="Wallet", icon=ft.Icons.ACCOUNT_BALANCE_WALLET, content=self.create_wallet_tab()),
                ft.Tab(text="Blockchain", icon=ft.Icons.LINK, content=self.create_blockchain_tab()),
                ft.Tab(text="Network", icon=ft.Icons.NETWORK_CHECK, content=self.create_network_tab()),
                ft.Tab(text="Mining", icon=ft.Icons.MEMORY, content=self.create_mining_tab()),
                ft.Tab(text="Settings", icon=ft.Icons.SETTINGS, content=self.create_settings_tab()),
                ft.Tab(text="Debug", icon=ft.Icons.BUG_REPORT, content=self.create_debug_tab()),
            ],
            expand=True
        )
        
        # Main container
        main_container = ft.Column([
            self.create_header(),
            self.tabs,
            self.status_bar
        ], expand=True, spacing=0)
        
        self.page.add(main_container)
    
    def create_header(self):
        """Create application header"""
        return ft.Container(
            content=ft.Row([
                ft.Icon(ft.Icons.CURRENCY_BITCOIN, size=40, color=ft.Colors.ORANGE),
                ft.Column([
                    ft.Text("Qbitcoin Node Manager", size=24, weight=ft.FontWeight.BOLD),
                    ft.Text("Post-Quantum Blockchain Node", size=12, color=ft.Colors.GREY)
                ], spacing=0),
                ft.Container(expand=True),
                self.create_node_controls()
            ], alignment=ft.MainAxisAlignment.SPACE_BETWEEN),
            padding=20,
            bgcolor=ft.Colors.BLUE_GREY_900,
            margin=0
        )
    
    def create_node_controls(self):
        """Create node control buttons"""
        self.start_button = ft.ElevatedButton(
            "Start Node",
            icon=ft.Icons.PLAY_ARROW,
            color=ft.Colors.WHITE,
            bgcolor=ft.Colors.GREEN,
            on_click=self.start_node
        )
        
        self.stop_button = ft.ElevatedButton(
            "Stop Node",
            icon=ft.Icons.STOP,
            color=ft.Colors.WHITE,
            bgcolor=ft.Colors.RED,
            on_click=self.stop_node,
            disabled=True
        )
        
        return ft.Row([self.start_button, self.stop_button], spacing=10)
    
    def create_dashboard_tab(self):
        """Create dashboard tab content"""
        # Node status cards
        self.node_status_card = self.create_status_card("Node Status", "Stopped", ft.Colors.RED)
        self.wallet_status_card = self.create_status_card("Wallet Status", "Locked", ft.Colors.ORANGE)
        self.blockchain_status_card = self.create_status_card("Blockchain Height", "0", ft.Colors.BLUE)
        self.network_status_card = self.create_status_card("Connected Peers", "0", ft.Colors.PURPLE)
        
        # Quick stats
        self.balance_card = self.create_info_card("Total Balance", "0.000000000 QBIT")
        self.hashrate_card = self.create_info_card("Mining Hashrate", "0 H/s")
        self.last_block_card = self.create_info_card("Last Block Time", "N/A")
        self.difficulty_card = self.create_info_card("Network Difficulty", "N/A")
        
        # Recent activity
        self.activity_list = ft.ListView(
            height=200,
            spacing=5,
            auto_scroll=True
        )
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Dashboard", size=20, weight=ft.FontWeight.BOLD),
                ft.Row([
                    self.node_status_card,
                    self.wallet_status_card,
                    self.blockchain_status_card,
                    self.network_status_card
                ], spacing=10),
                ft.Row([
                    self.balance_card,
                    self.hashrate_card,
                    self.last_block_card,
                    self.difficulty_card
                ], spacing=10),
                ft.Container(
                    content=ft.Column([
                        ft.Text("Recent Activity", size=16, weight=ft.FontWeight.BOLD),
                        self.activity_list
                    ]),
                    bgcolor=ft.Colors.GREY_800,  # Dark container
                    padding=15,
                    border_radius=10,
                    expand=True
                )
            ], spacing=15),
            padding=20,
            expand=True
        )
    
    def create_wallet_tab(self):
        """Create wallet tab content"""
        # Wallet controls
        self.wallet_unlock_btn = ft.ElevatedButton(
            "Unlock Wallet",
            icon=ft.Icons.LOCK_OPEN,
            on_click=self.unlock_wallet_dialog
        )
        
        self.wallet_create_btn = ft.ElevatedButton(
            "Create Wallet",
            icon=ft.Icons.ADD,
            on_click=self.create_wallet_dialog
        )
        
        self.wallet_backup_btn = ft.ElevatedButton(
            "Backup Wallet",
            icon=ft.Icons.BACKUP,
            on_click=self.backup_wallet_dialog
        )
        
        # Address list
        self.address_list = ft.DataTable(
            columns=[
                ft.DataColumn(ft.Text("Address")),
                ft.DataColumn(ft.Text("Balance")),
                ft.DataColumn(ft.Text("Type")),
                ft.DataColumn(ft.Text("Actions"))
            ],
            rows=[]
        )
        
        # Refresh button for wallet addresses
        self.refresh_wallet_btn = ft.IconButton(
            icon=ft.Icons.REFRESH,
            tooltip="Refresh wallet addresses",
            on_click=self.load_wallet_addresses
        )
        
        # Transaction form
        self.send_to_field = ft.TextField(
            label="Send to Address",
            width=400,
            prefix_icon=ft.Icons.ACCOUNT_BALANCE_WALLET
        )
        
        self.send_amount_field = ft.TextField(
            label="Amount (QBIT)",
            width=200,
            prefix_icon=ft.Icons.MONEY
        )
        
        self.send_fee_field = ft.TextField(
            label="Fee",
            width=150,
            value="0.01",
            prefix_icon=ft.Icons.PAID
        )
        
        self.send_button = ft.ElevatedButton(
            "Send Transaction",
            icon=ft.Icons.SEND,
            on_click=self.send_transaction
        )
        
        # Load wallet addresses when the tab is created
        self.load_wallet()
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Wallet Management", size=20, weight=ft.FontWeight.BOLD),
                
                # Wallet status card
                ft.Container(
                    content=ft.Row([
                        self.create_status_card("Wallet Status", "Not Loaded", ft.Colors.ORANGE),
                        self.create_status_card("Total Balance", "0.0 QBIT", ft.Colors.BLUE),
                        self.refresh_wallet_btn
                    ]),
                    padding=5
                ),
                
                # Wallet controls
                ft.Container(
                    content=ft.Column([
                        ft.Text("Wallet Controls", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([
                            self.wallet_unlock_btn,
                            self.wallet_create_btn,
                            self.wallet_backup_btn
                        ], spacing=10)
                    ]),
                    bgcolor=ft.Colors.GREY_800,  # Dark container
                    padding=15,
                    border_radius=10
                ),
                
                # Address list
                ft.Container(
                    content=ft.Column([
                        ft.Row([
                            ft.Text("Addresses", size=16, weight=ft.FontWeight.BOLD),
                            ft.IconButton(
                                icon=ft.Icons.REFRESH, 
                                tooltip="Refresh addresses",
                                on_click=self.load_wallet_addresses
                            )
                        ]),
                        ft.ListView(
                            controls=[self.address_list],
                            height=200
                        )
                    ]),
                    bgcolor=ft.Colors.GREY_800,  # Dark container
                    padding=15,
                    border_radius=10
                ),
                
                # Send transaction
                ft.Container(
                    content=ft.Column([
                        ft.Text("Send Transaction", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([
                            self.send_to_field,
                            self.send_amount_field,
                            self.send_fee_field,
                            self.send_button
                        ], spacing=10, wrap=True)
                    ]),
                    bgcolor=ft.Colors.GREY_800,  # Dark container
                    padding=15,
                    border_radius=10
                )
            ], spacing=15),
            padding=20,
            expand=True
        )
    
    def create_blockchain_tab(self):
        """Create blockchain tab content"""
        # Block explorer
        self.block_search_field = ft.TextField(
            label="Search by Block Height or Hash",
            width=300,
            prefix_icon=ft.Icons.SEARCH
        )
        
        self.block_search_btn = ft.ElevatedButton(
            "Search",
            icon=ft.Icons.SEARCH,
            on_click=self.search_block
        )
        
        # Block list
        self.block_list = ft.DataTable(
            columns=[
                ft.DataColumn(ft.Text("Height")),
                ft.DataColumn(ft.Text("Hash")),
                ft.DataColumn(ft.Text("Timestamp")),
                ft.DataColumn(ft.Text("Transactions")),
                ft.DataColumn(ft.Text("Size")),
                ft.DataColumn(ft.Text("Actions"))
            ],
            rows=[]
        )
        
        # Transaction pool
        self.mempool_list = ft.DataTable(
            columns=[
                ft.DataColumn(ft.Text("Transaction Hash")),
                ft.DataColumn(ft.Text("From")),
                ft.DataColumn(ft.Text("To")),
                ft.DataColumn(ft.Text("Amount")),
                ft.DataColumn(ft.Text("Fee"))
            ],
            rows=[]
        )
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Blockchain Explorer", size=20, weight=ft.FontWeight.BOLD),
                
                # Search
                ft.Container(
                    content=ft.Column([
                        ft.Text("Block Search", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([self.block_search_field, self.block_search_btn], spacing=10)
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Recent blocks
                ft.Container(
                    content=ft.Column([
                        ft.Text("Recent Blocks", size=16, weight=ft.FontWeight.BOLD),
                        ft.ListView(
                            controls=[self.block_list],
                            height=250
                        )
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Memory pool
                ft.Container(
                    content=ft.Column([
                        ft.Text("Memory Pool (Pending Transactions)", size=16, weight=ft.FontWeight.BOLD),
                        ft.ListView(
                            controls=[self.mempool_list],
                            height=200
                        )
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                )
            ], spacing=15, scroll=ft.ScrollMode.AUTO),
            padding=20,
            expand=True
        )
    
    def create_network_tab(self):
        """Create network tab content"""
        # Peer list
        self.peer_list = ft.DataTable(
            columns=[
                ft.DataColumn(ft.Text("IP Address")),
                ft.DataColumn(ft.Text("Port")),
                ft.DataColumn(ft.Text("Version")),
                ft.DataColumn(ft.Text("Height")),
                ft.DataColumn(ft.Text("Ping")),
                ft.DataColumn(ft.Text("Actions"))
            ],
            rows=[]
        )
        
        # Add peer controls
        self.add_peer_field = ft.TextField(
            label="Peer IP:Port",
            width=200,
            prefix_icon=ft.Icons.COMPUTER
        )
        
        self.add_peer_btn = ft.ElevatedButton(
            "Add Peer",
            icon=ft.Icons.ADD,
            on_click=self.add_peer
        )
        
        # Network stats
        self.network_stats = ft.Column([
            ft.Text("Network Statistics", size=16, weight=ft.FontWeight.BOLD),
            ft.Text("Loading...", size=12)
        ])
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Network Management", size=20, weight=ft.FontWeight.BOLD),
                
                # Network stats
                ft.Container(
                    content=self.network_stats,
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Add peer
                ft.Container(
                    content=ft.Column([
                        ft.Text("Add Peer", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([self.add_peer_field, self.add_peer_btn], spacing=10)
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Peer list
                ft.Container(
                    content=ft.Column([
                        ft.Text("Connected Peers", size=16, weight=ft.FontWeight.BOLD),
                        ft.ListView(
                            controls=[self.peer_list],
                            height=300
                        )
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                )
            ], spacing=15, scroll=ft.ScrollMode.AUTO),
            padding=20,
            expand=True
        )
    
    def create_mining_tab(self):
        """Create mining tab content"""
        # Mining controls
        self.mining_address_field = ft.TextField(
            label="Mining Address",
            width=400,
            value=self.settings.get("mining_address", ""),
            prefix_icon=ft.Icons.ACCOUNT_BALANCE_WALLET
        )
        
        self.mining_threads_field = ft.TextField(
            label="Threads",
            width=100,
            value=str(self.settings.get("mining_threads", 1)),
            prefix_icon=ft.Icons.MEMORY
        )
        
        self.start_mining_btn = ft.ElevatedButton(
            "Start Mining",
            icon=ft.Icons.PLAY_ARROW,
            color=ft.Colors.WHITE,
            bgcolor=ft.Colors.GREEN,
            on_click=self.start_mining
        )
        
        self.stop_mining_btn = ft.ElevatedButton(
            "Stop Mining",
            icon=ft.Icons.STOP,
            color=ft.Colors.WHITE,
            bgcolor=ft.Colors.RED,
            on_click=self.stop_mining,
            disabled=True
        )
        
        # Mining stats
        self.mining_stats = ft.Column([
            ft.Text("Mining Statistics", size=16, weight=ft.FontWeight.BOLD),
            ft.Text("Not mining", size=12)
        ])
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Mining Configuration", size=20, weight=ft.FontWeight.BOLD),
                
                # Mining controls
                ft.Container(
                    content=ft.Column([
                        ft.Text("Mining Settings", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([
                            self.mining_address_field,
                            self.mining_threads_field
                        ], spacing=10),
                        ft.Row([
                            self.start_mining_btn,
                            self.stop_mining_btn
                        ], spacing=10)
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Mining stats
                ft.Container(
                    content=self.mining_stats,
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10,
                    expand=True
                )
            ], spacing=15, scroll=ft.ScrollMode.AUTO),
            padding=20,
            expand=True
        )
    
    def create_settings_tab(self):
        """Create settings tab content"""
        # General settings
        self.auto_start_check = ft.Checkbox(
            label="Auto-start node on GUI launch",
            value=self.settings.get("auto_start", False)
        )
        
        self.data_dir_field = ft.TextField(
            label="Data Directory",
            width=400,
            value=self.settings.get("node_data_dir", config.user.qrl_dir)
        )
        
        # API settings
        self.api_enabled_check = ft.Checkbox(
            label="Enable API Server",
            value=self.settings.get("api_enabled", True)
        )
        
        self.api_port_field = ft.TextField(
            label="API Port",
            width=100,
            value=str(self.settings.get("api_port", 19009))
        )
        
        # P2P settings
        self.p2p_port_field = ft.TextField(
            label="P2P Port",
            width=100,
            value=str(self.settings.get("p2p_port", 19000))
        )
        
        # Log level
        self.log_level_dropdown = ft.Dropdown(
            label="Log Level",
            width=150,
            value=self.settings.get("log_level", "INFO"),
            options=[
                ft.dropdown.Option("DEBUG"),
                ft.dropdown.Option("INFO"),
                ft.dropdown.Option("WARNING"),
                ft.dropdown.Option("ERROR"),
                ft.dropdown.Option("CRITICAL")
            ]
        )
        
        self.save_settings_btn = ft.ElevatedButton(
            "Save Settings",
            icon=ft.Icons.SAVE,
            on_click=self.save_settings_action
        )
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Application Settings", size=20, weight=ft.FontWeight.BOLD),
                
                # General settings
                ft.Container(
                    content=ft.Column([
                        ft.Text("General", size=16, weight=ft.FontWeight.BOLD),
                        self.auto_start_check,
                        self.data_dir_field,
                        self.log_level_dropdown
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Network settings
                ft.Container(
                    content=ft.Column([
                        ft.Text("Network", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([
                            self.p2p_port_field,
                            self.api_enabled_check,
                            self.api_port_field
                        ], spacing=10)
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Save button
                ft.Container(
                    content=self.save_settings_btn,
                    alignment=ft.alignment.center
                )
            ], spacing=15, scroll=ft.ScrollMode.AUTO),
            padding=20,
            expand=True
        )
    
    def create_debug_tab(self):
        """Create debug tab content"""
        # Log viewer was already initialized in __init__
        # Just configure the rest of the tab
        
        # Log controls
        self.clear_logs_btn = ft.ElevatedButton(
            "Clear Logs",
            icon=ft.Icons.CLEAR,
            on_click=self.clear_logs
        )
        
        self.export_logs_btn = ft.ElevatedButton(
            "Export Logs",
            icon=ft.Icons.DOWNLOAD,
            on_click=self.export_logs
        )
        
        # Console input
        self.console_input = ft.TextField(
            label="Console Command",
            width=400,
            on_submit=self.execute_console_command
        )
        
        self.execute_btn = ft.ElevatedButton(
            "Execute",
            icon=ft.Icons.PLAY_ARROW,
            on_click=self.execute_console_command
        )
        
        return ft.Container(
            content=ft.Column([
                ft.Text("Debug Console", size=20, weight=ft.FontWeight.BOLD),
                
                # Log controls
                ft.Container(
                    content=ft.Row([
                        self.clear_logs_btn,
                        self.export_logs_btn
                    ], spacing=10),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Log viewer
                ft.Container(
                    content=ft.Column([
                        ft.Text("System Logs", size=16, weight=ft.FontWeight.BOLD),
                        ft.Container(
                            content=self.log_viewer,
                            bgcolor=ft.Colors.BLACK,
                            padding=10,
                            border_radius=5
                        )
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                ),
                
                # Console
                ft.Container(
                    content=ft.Column([
                        ft.Text("Console", size=16, weight=ft.FontWeight.BOLD),
                        ft.Row([
                            self.console_input,
                            self.execute_btn
                        ], spacing=10)
                    ]),
                    bgcolor=ft.Colors.GREY_800,
                    padding=15,
                    border_radius=10
                )
            ], spacing=15, scroll=ft.ScrollMode.AUTO),
            padding=20,
            expand=True,
        )
    
    def initialize_modals(self):
        """Initialize modal dialog instances"""
        try:
            self.block_details_modal = BlockDetailsModal(self.page)
            self.transaction_details_modal = TransactionDetailsModal(self.page)
            self.wallet_details_modal = WalletDetailsModal(self.page)
            self.add_log("Modal dialogs initialized successfully")
        except Exception as ex:
            self.add_log(f"Error initializing modals: {str(ex)}", "ERROR")
    
    def show_block_details(self, block_info):
        """Show detailed block information in modal"""
        try:
            if not self.block_details_modal:
                self.initialize_modals()
            
            if self.block_details_modal:
                self.block_details_modal.show(block_info)
                self.add_log(f"Showing details for block #{block_info.header.block_number}")
        except Exception as ex:
            self.add_log(f"Error showing block details: {str(ex)}", "ERROR")
    
    def show_transaction_details(self, transaction):
        """Show detailed transaction information in modal"""
        try:
            if not self.transaction_details_modal:
                self.initialize_modals()
            
            if self.transaction_details_modal:
                self.transaction_details_modal.show(transaction)
                self.add_log("Showing transaction details")
        except Exception as ex:
            self.add_log(f"Error showing transaction details: {str(ex)}", "ERROR")
    
    def show_wallet_details(self, wallet_info):
        """Show detailed wallet information in modal"""
        try:
            if not self.wallet_details_modal:
                self.initialize_modals()
            
            if self.wallet_details_modal:
                self.wallet_details_modal.show(wallet_info)
                self.add_log("Showing wallet details")
        except Exception as ex:
            self.add_log(f"Error showing wallet details: {str(ex)}", "ERROR")

    # Helper methods for UI components
    def create_status_card(self, title: str, value: str, color):
        """Create a status card"""
        return ft.Container(
            content=ft.Column([
                ft.Text(title, size=12, color=ft.Colors.GREY),
                ft.Text(value, size=18, weight=ft.FontWeight.BOLD, color=color)
            ], spacing=5, alignment=ft.MainAxisAlignment.CENTER),
            bgcolor=ft.Colors.GREY_800,
            padding=15,
            border_radius=10,
            width=200,
            height=80
        )
    
    def create_info_card(self, title: str, value: str):
        """Create an info card"""
        return ft.Container(
            content=ft.Column([
                ft.Text(title, size=12, color=ft.Colors.GREY),
                ft.Text(value, size=14, weight=ft.FontWeight.BOLD)
            ], spacing=5, alignment=ft.MainAxisAlignment.CENTER),
            bgcolor=ft.Colors.GREY_800,
            padding=15,
            border_radius=10,
            width=200,
            height=80
        )
    
    # Event handlers
    def start_node(self, e):
        """Start the Qbitcoin node using main()"""
        try:
            self.add_log("Starting Qbitcoin node...")
            
            # Update UI state
            self.start_button.disabled = True
            self.stop_button.disabled = False
            self.update_status("Node Starting...", ft.Colors.ORANGE)
            self.page.update()
            
            def startup_thread():
                try:
                    # Set mining address in config if provided
                    if self.settings.get("mining_address"):
                        mining_addr_str = self.settings["mining_address"]
                        self.add_log(f"Using mining address: {mining_addr_str}")
                    
                    # Import and call main() directly - same as start_qbitcoin.py
                    from qbitcoin.main import main
                    
                    self.add_log("Starting Qbitcoin node via main()...")
                    
                    # Start the node in a subprocess to avoid blocking
                    import sys
                    import subprocess
                    
                    # Build command line arguments for main()
                    cmd = [sys.executable, "-c", """
import os
import sys
sys.path.insert(0, os.path.abspath('.'))
os.environ['PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION'] = 'python'
from qbitcoin.main import main
main()
"""]
                    
                    # Start the node process
                    self.node_process = subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        universal_newlines=True,
                        cwd=os.path.dirname(__file__)
                    )
                    
                    self.add_log("Node process started")
                    
                    # Give it a moment to start
                    time.sleep(3)
                    
                    # Try to connect to gRPC services
                    if self.connect_grpc_services():
                        self.node_running = True
                        self.update_status("Node Running", ft.Colors.GREEN)
                        self.add_log("Qbitcoin node started successfully")
                    else:
                        self.add_log("Node started but gRPC connection failed", "WARNING")
                        self.node_running = True
                        self.update_status("Node Running (API pending)", ft.Colors.ORANGE)
                    
                    self.page.update()
                    
                    # Monitor node process output
                    def monitor_output():
                        if self.node_process:
                            for line in iter(self.node_process.stdout.readline, ''):
                                if line:
                                    # Clean up the log line and add to GUI
                                    cleaned_line = line.strip()
                                    if cleaned_line:
                                        self.add_log(f"Node: {cleaned_line}")
                                if self.node_process.poll() is not None:
                                    break
                    
                    # Start output monitoring in background
                    threading.Thread(target=monitor_output, daemon=True).start()
                    
                except Exception as ex:
                    self.add_log(f"Error starting node: {str(ex)}", "ERROR")
                    import traceback
                    self.add_log(f"Traceback: {traceback.format_exc()}", "ERROR")
                    self.update_status("Start Failed", ft.Colors.RED)
                    self.start_button.disabled = False
                    self.stop_button.disabled = True
                    self.page.update()
            
            threading.Thread(target=startup_thread, daemon=True).start()
            
        except Exception as ex:
            self.add_log(f"Error starting node: {str(ex)}", "ERROR")
            self.update_status("Start Failed", ft.Colors.RED)
            self.start_button.disabled = False
            self.stop_button.disabled = True
            self.page.update()
    
    def stop_node(self, e):
        """Stop the Qbitcoin node"""
        try:
            self.add_log("Stopping Qbitcoin node...")
            
            # Stop node process
            if self.node_process:
                self.add_log("Terminating node process...")
                self.node_process.terminate()
                
                # Wait for process to terminate gracefully
                try:
                    self.node_process.wait(timeout=10)
                except subprocess.TimeoutExpired:
                    self.add_log("Force killing node process...")
                    self.node_process.kill()
                    self.node_process.wait()
                
                self.node_process = None
            
            # Reset connection references
            self.public_stub = None
            self.wallet_stub = None
            self.mining_stub = None
            
            # Update UI state
            self.node_running = False
            self.start_button.disabled = False
            self.stop_button.disabled = True
            self.start_mining_btn.disabled = True
            self.stop_mining_btn.disabled = True
            
            self.update_status("Node Stopped", ft.Colors.RED)
            self.add_log("Qbitcoin node stopped successfully")
            self.page.update()
            
        except Exception as ex:
            self.add_log(f"Error stopping node: {str(ex)}", "ERROR")
    
    def start_mining(self, e):
        """Start mining"""
        try:
            mining_address = self.mining_address_field.value
            threads = int(self.mining_threads_field.value)
            
            if not mining_address:
                self.show_error_dialog("Mining address is required")
                return
            
            if not self.node_running:
                self.show_error_dialog("Node must be running to start mining")
                return
            
            # Validate mining address
            try:
                # Convert Q address to bytes for validation
                if not mining_address.startswith('Q'):
                    self.show_error_dialog("Invalid mining address format")
                    return
            except Exception:
                self.show_error_dialog("Invalid mining address")
                return
            
            # Since we're using main() directly, mining would be controlled via the node's API
            # For now, we'll just update the UI to indicate mining is "started"
            # In a real implementation, you'd use the mining API to start/stop mining
            
            # Update UI
            self.start_mining_btn.disabled = True
            self.stop_mining_btn.disabled = False
            
            # Update settings
            self.settings["mining_address"] = mining_address
            self.settings["mining_threads"] = threads
            self.settings["mining_enabled"] = True
            self.save_settings()
            
            self.add_log(f"Mining configured for {threads} threads with address {mining_address}")
            self.add_log("Note: Mining control via API will be implemented in future updates")
            self.page.update()
            
        except Exception as ex:
            self.add_log(f"Error starting mining: {str(ex)}", "ERROR")
    
    def stop_mining(self, e):
        """Stop mining"""
        try:
            # Since we're using main() directly, mining would be controlled via the node's API
            # For now, we'll just update the UI to indicate mining is "stopped"
            
            # Update UI
            self.start_mining_btn.disabled = False
            self.stop_mining_btn.disabled = True
            
            # Update settings
            self.settings["mining_enabled"] = False
            self.save_settings()
            
            self.add_log("Mining stopped")
            self.add_log("Note: Mining control via API will be implemented in future updates")
            self.page.update()
            
        except Exception as ex:
            self.add_log(f"Error stopping mining: {str(ex)}", "ERROR")
    
    def unlock_wallet_dialog(self, e):
        """Show wallet unlock dialog"""
        from qbitcoin.core.Wallet import Wallet, WalletDecryptionError
        
        def unlock_wallet(e):
            password = password_field.value
            if password:
                try:
                    if not self.wallet:
                        self.wallet = Wallet(wallet_path=self.wallet_path)
                        self.wallet_loaded = True
                    
                    # Try to decrypt the wallet with provided password
                    if self.wallet.encrypted:
                        self.wallet.decrypt(password)
                        self.wallet_unlocked = True
                        self.update_wallet_status("Unlocked", ft.Colors.GREEN)
                        self.add_log("Wallet unlocked successfully")
                        
                        # Reload addresses with updated wallet
                        self.load_wallet_addresses()
                    else:
                        self.add_log("Wallet is not encrypted", "INFO")
                        
                    dialog.open = False
                    self.page.update()
                except WalletDecryptionError as ex:
                    self.add_log("Invalid password. Please try again.", "ERROR")
                    error_text.value = "Invalid password. Please try again."
                    self.page.update()
                except Exception as ex:
                    self.add_log(f"Error unlocking wallet: {str(ex)}", "ERROR")
                    error_text.value = f"Error: {str(ex)}"
                    self.page.update()
        
        password_field = ft.TextField(
            label="Password",
            password=True,
            autofocus=True,
            on_submit=lambda e: unlock_wallet(e)
        )
        
        error_text = ft.Text("", color=ft.Colors.RED_500)
        
        dialog = ft.AlertDialog(
            modal=True,
            title=ft.Text("Unlock Wallet"),
            content=ft.Column([
                ft.Text("Enter your wallet password:"),
                password_field,
                error_text
            ], height=100, spacing=10),
            actions=[
                ft.TextButton("Cancel", on_click=lambda e: setattr(dialog, 'open', False)),
                ft.TextButton("Unlock", on_click=unlock_wallet)
            ]
        )
        
        self.page.dialog = dialog
        dialog.open = True
        self.page.update()
    
    def create_wallet_dialog(self, e):
        """Show create wallet dialog"""
        from qbitcoin.core.Wallet import Wallet, WalletException
        
        def create_wallet(e):
            name = name_field.value
            password = password_field.value
            
            if not name:
                error_text.value = "Wallet name is required"
                error_text.visible = True
                self.page.update()
                return
            
            try:
                # Create new Falcon-512 wallet
                self.wallet = Wallet(wallet_path=self.wallet_path)
                
                # Check if wallet already exists and has addresses
                if len(self.wallet.address_items) > 0:
                    error_text.value = "A wallet already exists. Use it or delete it before creating a new one."
                    error_text.visible = True
                    self.page.update()
                    return
                
                # Add a new Falcon-512 address
                falcon_keypair = self.wallet.add_new_address()
                
                # Encrypt if password provided
                if password:
                    self.wallet.encrypt(password)
                    self.add_log("Wallet encrypted with provided password")
                
                # Save wallet
                self.wallet.save()
                self.wallet_exists = True
                self.wallet_loaded = True
                
                if not password:
                    self.wallet_unlocked = True
                    self.update_wallet_status("Unlocked", ft.Colors.GREEN)
                else:
                    self.update_wallet_status("Locked", ft.Colors.ORANGE)
                
                # Show success message with the new address
                self.add_log(f"Created new wallet with address: {falcon_keypair['address']}")
                dialog.open = False
                self.page.update()
                
                # Load addresses to update the UI
                self.load_wallet_addresses()
            except Exception as ex:
                self.add_log(f"Error creating wallet: {str(ex)}", "ERROR")
                error_text.value = f"Error: {str(ex)}"
                error_text.visible = True
                self.page.update()
        
        name_field = ft.TextField(
            label="Wallet Name", 
            autofocus=True,
            hint_text="Enter a name for your wallet",
            width=350
        )
        
        password_field = ft.TextField(
            label="Password (Optional)", 
            password=True,
            hint_text="Leave empty for no encryption",
            width=350
        )
        
        confirm_password = ft.TextField(
            label="Confirm Password", 
            password=True,
            width=350
        )
        
        error_text = ft.Text("", color=ft.Colors.RED_500, visible=False)
        
        dialog = ft.AlertDialog(
            modal=True,
            title=ft.Text("Create New Wallet"),
            content=ft.Column([
                ft.Text("This will create a new Falcon-512 post-quantum wallet"),
                name_field,
                password_field,
                confirm_password,
                error_text
            ], height=220, spacing=10),
            actions=[
                ft.TextButton("Cancel", on_click=lambda e: setattr(dialog, 'open', False)),
                ft.TextButton("Create", on_click=create_wallet)
            ]
        )
        
        self.page.dialog = dialog
        dialog.open = True
        self.page.update()
    
    def backup_wallet_dialog(self, e):
        """Show backup wallet dialog"""
        self.show_info_dialog("Backup wallet feature will be implemented here")
    
    def send_transaction(self, e):
        """Send a transaction"""
        try:
            to_address = self.send_to_field.value
            amount = self.send_amount_field.value
            fee = self.send_fee_field.value
            
            if not to_address or not amount:
                self.show_error_dialog("To address and amount are required")
                return
            
            # Validate and send transaction
            self.add_log(f"Sending {amount} QBIT to {to_address} with fee {fee}")
            
            # Clear fields
            self.send_to_field.value = ""
            self.send_amount_field.value = ""
            self.page.update()
            
        except Exception as ex:
            self.add_log(f"Error sending transaction: {str(ex)}", "ERROR")
    
    def search_block(self, e):
        """Search for a block"""
        search_term = self.block_search_field.value
        if not search_term:
            return
            
        try:
            self.add_log(f"Searching for block: {search_term}")
            
            if not self.node_running or not self.public_stub:
                self.show_error_dialog("Node must be running and connected to search blocks")
                return
            
            # Clear existing block list
            self.block_list.rows.clear()
            
            # Try to parse as block number first, then as hash
            block_info = None
            found = False
            
            # First try: parse as block number
            try:
                block_number = int(search_term)
                self.add_log(f"Searching for block number: {block_number}")
                request = qbit_pb2.GetBlockByNumberReq(block_number=block_number)
                response = self.public_stub.GetBlockByNumber(request, timeout=10)
                block_info = response.block
                self.add_log(f"Found block #{block_number}")
                found = True
                
            except ValueError:
                # Not a valid integer, try as block hash
                self.add_log("Not a valid block number, trying as hash...")
                
            except grpc.RpcError as grpc_ex:
                self.add_log(f"gRPC error getting block by number: {grpc_ex}", "ERROR")
                # Don't return here, might still be a valid hash
                
            # Second try: parse as block hash (only if not found as number)
            if not found:
                try:
                    # Convert hex string to bytes
                    hash_term = search_term
                    if hash_term.startswith('0x'):
                        hash_term = hash_term[2:]
                    
                    # Validate hex format
                    block_hash = bytes.fromhex(hash_term)
                    self.add_log(f"Searching for block hash: {hash_term}")
                    request = qbit_pb2.GetBlockReq(header_hash=block_hash)
                    response = self.public_stub.GetBlock(request, timeout=10)
                    block_info = response.block
                    self.add_log(f"Found block by hash")
                    found = True
                    
                except ValueError as hex_ex:
                    self.add_log(f"Invalid hex format: {hex_ex}", "ERROR")
                    
                except grpc.RpcError as grpc_ex:
                    self.add_log(f"gRPC error getting block by hash: {grpc_ex}", "ERROR")
                    
            if found and block_info:
                # Display block information
                self.display_block_info(block_info)
                self.add_log("Block information displayed")
            else:
                self.show_error_dialog("Block not found")
                
        except Exception as ex:
            self.add_log(f"Error searching for block: {str(ex)}", "ERROR")
            self.show_error_dialog(f"Error searching for block: {str(ex)}")
             
            
            if block_info and found:
                # Display block information
                self.display_block_info(block_info)
                self.add_log("Block information displayed")
            else:
                self.show_error_dialog("Block not found or invalid search term")
                self.add_log("Block not found", "WARNING")
                
        except Exception as ex:
            self.add_log(f"Error searching for block: {str(ex)}", "ERROR")
            self.show_error_dialog(f"Error searching for block: {str(ex)}")
    
    def display_block_info(self, block_info):
        """Display block information in the block list"""
        try:
            # Extract block data
            block_number = block_info.header.block_number
            block_hash = bin2hstr(block_info.header.hash_header)
            timestamp = block_info.header.timestamp_seconds
            tx_count = len(block_info.transactions)
            block_size = len(block_info.SerializeToString())
            
            # Convert timestamp to readable format
            readable_time = datetime.datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
            
            # Add special indicator for genesis block
            block_display = str(block_number)
            if block_number == 0:
                block_display = "0 (Genesis)"
            
            # Create "View Details" button
            view_button = ft.ElevatedButton(
                "View Details",
                on_click=lambda e: self.show_block_details(block_info),
                style=ft.ButtonStyle(
                    bgcolor=ft.Colors.BLUE_600,
                    color=ft.Colors.WHITE,
                    padding=ft.Padding(8, 4, 8, 4)
                ),
                height=30
            )
            
            # Add to block list
            self.block_list.rows.append(
                ft.DataRow(
                    cells=[
                        ft.DataCell(ft.Text(block_display)),
                        ft.DataCell(ft.Text(f"{block_hash[:16]}..." if len(block_hash) > 16 else block_hash)),
                        ft.DataCell(ft.Text(readable_time)),
                        ft.DataCell(ft.Text(str(tx_count))),
                        ft.DataCell(ft.Text(f"{block_size:,} bytes")),
                        ft.DataCell(view_button)
                    ]
                )
            )
            
            # Also show detailed block information in logs
            if block_number == 0:
                self.add_log("=" * 50)
                self.add_log("GENESIS BLOCK FOUND!")
                self.add_log("=" * 50)
            
            self.add_log(f"Block #{block_number} Details:")
            self.add_log(f"  Hash: {block_hash}")
            self.add_log(f"  Timestamp: {readable_time}")
            self.add_log(f"  Transactions: {tx_count}")
            self.add_log(f"  Size: {block_size:,} bytes")
            
            if block_number == 0:
                self.add_log("  ** This is the Genesis Block **")
                # Show additional genesis block details
                self.add_log(f"  Previous Hash: {bin2hstr(block_info.header.hash_header_prev) if block_info.header.hash_header_prev else 'None (Genesis)'}")
                self.add_log(f"  Merkle Root: {bin2hstr(block_info.header.merkle_root) if block_info.header.merkle_root else 'N/A'}")
                
                # Show transaction details for genesis block
                if tx_count > 0:
                    self.add_log("  Genesis Block Transactions:")
                    for i, tx in enumerate(block_info.transactions):
                        tx_hash = bin2hstr(tx.transaction_hash) if tx.transaction_hash else "N/A"
                        self.add_log(f"    TX {i}: {tx_hash[:32]}...")
                        
                        # Show more details for first transaction
                        if i == 0:
                            tx_type = getattr(tx, 'type', 'unknown')
                            self.add_log(f"      Type: {tx_type}")
                            
            self.page.update()
            
        except Exception as ex:
            self.add_log(f"Error displaying block info: {str(ex)}", "ERROR")
            
            # Show transaction details for genesis block
            if block_number == 0 and tx_count > 0:
                self.add_log("  Genesis Block Transactions:")
                for i, tx in enumerate(block_info.transactions):
                    tx_hash = bin2hstr(tx.transaction_hash) if tx.transaction_hash else "N/A"
                    self.add_log(f"    TX {i}: {tx_hash[:16]}...")
            
            self.page.update()
            
        except Exception as ex:
            self.add_log(f"Error displaying block info: {str(ex)}", "ERROR")
    
    def add_peer(self, e):
        """Add a new peer"""
        peer_address = self.add_peer_field.value
        if peer_address:
            self.add_log(f"Adding peer: {peer_address}")
            self.add_peer_field.value = ""
            self.page.update()
    
    def save_settings_action(self, e):
        """Save application settings"""
        try:
            self.settings.update({
                "auto_start": self.auto_start_check.value,
                "node_data_dir": self.data_dir_field.value,
                "api_enabled": self.api_enabled_check.value,
                "api_port": int(self.api_port_field.value),
                "p2p_port": int(self.p2p_port_field.value),
                "log_level": self.log_level_dropdown.value,
                "mining_address": self.mining_address_field.value,
                "mining_threads": int(self.mining_threads_field.value)
            })
            
            self.save_settings()
            self.add_log("Settings saved successfully")
            
        except Exception as ex:
            self.add_log(f"Error saving settings: {str(ex)}", "ERROR")
    
    def clear_logs(self, e):
        """Clear debug logs"""
        self.log_viewer.controls.clear()
        self.debug_logs.clear()
        self.page.update()
    
    def export_logs(self, e):
        """Export logs to file"""
        try:
            timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            log_file = f"qbitcoin_logs_{timestamp}.txt"
            
            with open(log_file, 'w') as f:
                for log_entry in self.debug_logs:
                    f.write(f"{log_entry}\n")
            
            self.add_log(f"Logs exported to {log_file}")
            
        except Exception as ex:
            self.add_log(f"Error exporting logs: {str(ex)}", "ERROR")
    
    def execute_console_command(self, e):
        """Execute console command"""
        command = self.console_input.value
        if command:
            self.add_log(f"> {command}")
            # Process command here
            self.add_log("Command executed (feature in development)")
            self.console_input.value = ""
            self.page.update()
    
    # Utility methods
    def update_status(self, text: str, color):
        """Update status bar"""
        self.status_text.value = text
        self.status_text.color = color
    
    def add_log(self, message: str, level: str = "INFO"):
        """Add log entry"""
        timestamp = datetime.datetime.now().strftime("%H:%M:%S")
        log_entry = f"[{timestamp}] {level}: {message}"
        
        self.debug_logs.append(log_entry)
        
        # Color code by level
        color = ft.Colors.WHITE
        if level == "ERROR":
            color = ft.Colors.RED
        elif level == "WARNING":
            color = ft.Colors.ORANGE
        elif level == "DEBUG":
            color = ft.Colors.GREY
        
        log_control = ft.Text(
            log_entry,
            size=10,
            color=color,
            font_family="monospace"
        )
        
        self.log_viewer.controls.append(log_control)
        
        # Keep only last 1000 log entries
        if len(self.log_viewer.controls) > 1000:
            self.log_viewer.controls.pop(0)
            self.debug_logs.pop(0)
        
        if hasattr(self, 'page'):
            self.page.update()
    
    def show_error_dialog(self, message: str):
        """Show error dialog"""
        dialog = ft.AlertDialog(
            modal=True,
            title=ft.Text("Error"),
            content=ft.Text(message),
            actions=[ft.TextButton("OK", on_click=lambda e: setattr(dialog, 'open', False))]
        )
        self.page.dialog = dialog
        dialog.open = True
        self.page.update()
    
    def show_info_dialog(self, message: str):
        """Show info dialog"""
        dialog = ft.AlertDialog(
            modal=True,
            title=ft.Text("Information"),
            content=ft.Text(message),
            actions=[ft.TextButton("OK", on_click=lambda e: setattr(dialog, 'open', False))]
        )
        self.page.dialog = dialog
        dialog.open = True
        self.page.update()
    
    def start_monitoring(self):
        """Start background monitoring"""
        def monitor():
            while True:
                try:
                    if self.node_running:
                        # Update dashboard data
                        self.update_dashboard_data()
                    time.sleep(5)  # Update every 5 seconds
                except Exception as ex:
                    self.add_log(f"Monitor error: {str(ex)}", "ERROR")
                    time.sleep(10)
        
        threading.Thread(target=monitor, daemon=True).start()
    
    def update_dashboard_data(self):
        """Update dashboard with real-time data"""
        try:
            if not self.node_running or not self.public_stub:
                return
                
            # Load recent blocks into blockchain tab
            self.load_recent_blocks()
            
            # Update node status
            node_status = self.get_node_status()
            if node_status:
                # Update status display if needed
                pass
                
        except Exception as ex:
            self.add_log(f"Error updating dashboard: {str(ex)}", "ERROR")
    
    def setup_logging(self):
        """Setup logging to capture node logs in GUI"""
        try:
            # Create custom handler for GUI
            self.gui_log_handler = GUILogHandler(self)
            self.gui_log_handler.setLevel(logging.INFO)
            
            # Create formatter
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            self.gui_log_handler.setFormatter(formatter)
            
            # Add handler to QRL logger
            qrl_logger = logging.getLogger('qbitcoin')
            qrl_logger.addHandler(self.gui_log_handler)
            qrl_logger.setLevel(logging.INFO)
            
            # Also add to root logger to catch all logs
            root_logger = logging.getLogger()
            root_logger.addHandler(self.gui_log_handler)
            
            # Add to specific loggers that might be used
            for logger_name in ['qbitcoin.core', 'qbitcoin.services', 'qbitcoin.p2p']:
                specific_logger = logging.getLogger(logger_name)
                specific_logger.addHandler(self.gui_log_handler)
                specific_logger.setLevel(logging.INFO)
            
            self.add_log("Logging system initialized", "INFO")
            
        except Exception as e:
            print(f"Error setting up logging: {e}")
    
    def load_recent_blocks(self):
        """Load recent blocks into the blockchain tab"""
        try:
            if not self.node_running or not self.public_stub:
                return
            
            # Get current blockchain height
            node_state_request = qbit_pb2.GetNodeStateReq()
            node_state_response = self.public_stub.GetNodeState(node_state_request, timeout=5)
            current_height = node_state_response.info.block_height
            
            # Clear existing rows
            self.block_list.rows.clear()
            
            # Always try to load genesis block first (block 0)
            blocks_to_load = []
            
            # Add genesis block
            blocks_to_load.append(0)
            
            # Add recent blocks (last 9 blocks)
            if current_height > 0:
                start_block = max(1, current_height - 8)  # Skip genesis since we already added it
                for block_num in range(current_height, start_block - 1, -1):
                    if block_num not in blocks_to_load:
                        blocks_to_load.append(block_num)
            
            # Load all blocks
            for block_num in blocks_to_load:
                try:
                    request = qbit_pb2.GetBlockByNumberReq(block_number=block_num)
                    response = self.public_stub.GetBlockByNumber(request, timeout=5)
                    block_info = response.block
                    
                    if block_info:
                        # Extract block data
                        block_number = block_info.header.block_number
                        block_hash = bin2hstr(block_info.header.hash_header)
                        timestamp = block_info.header.timestamp_seconds
                        tx_count = len(block_info.transactions)
                        block_size = len(block_info.SerializeToString())
                        
                        # Convert timestamp to readable format
                        readable_time = datetime.datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
                        
                        # Add special indicator for genesis block
                        block_display = str(block_number)
                        if block_number == 0:
                            block_display = "0 (Genesis)"
                        
                        # Create view details button
                        view_details_btn = ft.ElevatedButton(
                            "View Details",
                            icon=ft.Icons.VISIBILITY,
                            style=ft.ButtonStyle(
                                color=ft.Colors.WHITE,
                                bgcolor=ft.Colors.BLUE,
                                padding=ft.padding.symmetric(horizontal=8, vertical=4)
                            ),
                            on_click=lambda e, block=block_info: self.show_block_details(block)
                        )
                        
                        # Add to block list
                        self.block_list.rows.append(
                            ft.DataRow(
                                cells=[
                                    ft.DataCell(ft.Text(block_display)),
                                    ft.DataCell(ft.Text(f"{block_hash[:16]}..." if len(block_hash) > 16 else block_hash)),
                                    ft.DataCell(ft.Text(readable_time)),
                                    ft.DataCell(ft.Text(str(tx_count))),
                                    ft.DataCell(ft.Text(f"{block_size:,} bytes")),
                                    ft.DataCell(view_details_btn)
                                ]
                            )
                        )
                        
                         
                except Exception as block_ex:
                    self.add_log(f"Error loading block {block_num}: {block_ex}", "WARNING")
                    continue
            
            # Update the page
            if hasattr(self, 'page'):
                self.page.update()
                
        except Exception as ex:
            self.add_log(f"Error loading recent blocks: {str(ex)}", "ERROR")
    
    def load_wallet(self):
        """Initialize or load wallet"""
        try:
            import os
            from qbitcoin.core.Wallet import Wallet
            
            # Define wallet path based on config
            self.wallet_path = os.path.join(config.user.wallet_dir, config.dev.wallet_dat_filename)
            self.wallet_exists = os.path.exists(self.wallet_path)
            self.wallet_loaded = False
            self.wallet_unlocked = False
            
            # Initialize wallet object (doesn't load yet)
            self.wallet = None
            
            # Show status
            if self.wallet_exists:
                self.add_log(f"Found wallet at: {self.wallet_path}")
                self.update_wallet_status("Found (Locked)", ft.Colors.ORANGE)
            else:
                self.add_log("No wallet found. Create a new wallet to get started.")
                self.update_wallet_status("Not Found", ft.Colors.RED)
                # Automatically show wallet creation dialog
                # Schedule it with a timer to ensure UI is ready first
                self.create_wallet_dialog(None)
                
        except Exception as ex:
            self.add_log(f"Error initializing wallet: {str(ex)}", "ERROR")
            self.update_wallet_status("Error", ft.Colors.RED)
            
    def load_wallet_addresses(self, e=None):
        """Load wallet addresses into UI"""
        try:
            from qbitcoin.core.Wallet import Wallet, WalletException
            import base64
            
            # Clear existing addresses
            self.address_list.rows.clear()
            
            # Check if wallet file exists
            if not self.wallet_exists:
                self.add_log("No wallet file found. Create a wallet first.", "WARNING")
                return
                
            # Initialize wallet if not already
            if not self.wallet:
                self.wallet = Wallet(wallet_path=self.wallet_path)
                self.wallet_loaded = True
                
            # Get all addresses from wallet
            total_balance = 0
            
            for idx, item in enumerate(self.wallet.address_items):
                try:
                    # Get address
                    address = item.qaddress
                    
                    # Get balance (mock for now, would normally come from node API)
                    balance = 0.0
                    try:
                        # If connected to node, try to get actual balance
                        if self.node_running and self.public_stub:
                            # Convert Q address to bytes
                            if address.startswith('Q'):
                                addr_bytes = bytes(hstr2bin(address[1:]))
                                request = qbit_pb2.GetBalanceReq(address=addr_bytes)
                                response = self.public_stub.GetBalance(request, timeout=5)
                                balance = response.balance / 10**9  # Convert quark to Qbitcoin
                                total_balance += balance
                    except:
                        # If error, use 0 balance
                        balance = 0.0
                    
                    # Create "View" button
                    view_button = ft.ElevatedButton(
                        "View",
                        style=ft.ButtonStyle(
                            color=ft.Colors.WHITE,
                            bgcolor=ft.Colors.BLUE,
                            padding=ft.padding.symmetric(horizontal=8, vertical=4)
                        ),
                        on_click=lambda e, addr=address: self.show_wallet_details({'address': addr})
                    )
                    
                    # Add to address list
                    self.address_list.rows.append(
                        ft.DataRow(
                            cells=[
                                ft.DataCell(ft.Text(f"{address[:20]}..." if len(address) > 20 else address)),
                                ft.DataCell(ft.Text(f"{balance:.9f} QBIT")),
                                ft.DataCell(ft.Text("Falcon-512")),
                                ft.DataCell(view_button)
                            ]
                        )
                    )
                except Exception as item_ex:
                    self.add_log(f"Error processing address {idx}: {str(item_ex)}", "WARNING")
                    continue
            
            # Update wallet stats
            self.update_wallet_status(
                "Loaded" if not self.wallet.encrypted else "Locked", 
                ft.Colors.GREEN if not self.wallet.encrypted else ft.Colors.ORANGE
            )
            
            # Update total balance
            self.update_total_balance(total_balance)
            
            # Update UI
            self.page.update()
            self.add_log(f"Loaded {len(self.wallet.address_items)} addresses from wallet")
            
        except Exception as ex:
            self.add_log(f"Error loading wallet addresses: {str(ex)}", "ERROR")
            self.update_wallet_status("Error", ft.Colors.RED)
            
    def update_wallet_status(self, status_text, color):
        """Update wallet status display"""
        # Update the wallet status card on dashboard
        self.wallet_status_card.content.controls[1].value = status_text
        self.wallet_status_card.content.controls[1].color = color
        
        # Update the page if it exists
        if hasattr(self, 'page'):
            self.page.update()
            
    def update_total_balance(self, balance):
        """Update total balance display"""
        # Format the balance with proper precision
        formatted_balance = f"{balance:.9f} QBIT"
        
        try:
            # Update the balance card
            self.balance_card.content.controls[1].value = formatted_balance
            
            # Update the page if it exists
            if hasattr(self, 'page'):
                self.page.update()
        except:
            pass

def main():
    """
    Main function to run the QbitcoinGUI application
    Supports both desktop and web browser modes
    """
    # Create config directories if they don't exist
    from qbitcoin.core import config
    import os
    
    os.makedirs(config.user.wallet_dir, exist_ok=True)
    os.makedirs(config.user.qrl_dir, exist_ok=True)
    
    # Run as web application by default
    ft.app(target=QbitcoinGUI, view=ft.AppView.WEB_BROWSER, port=8080, assets_dir="assets")

if __name__ == "__main__":
    main()




#!/usr/bin/env python3
import base64

falcon_address = 'Ac8oK2beuuYpbNQBfIrqmDpwfqFOYitCXA=='
decoded = base64.b64decode(falcon_address)
print(f'Falcon address length: {len(decoded)} bytes')
print(f'Falcon address hex: {decoded.hex()}')
print(f'First byte: 0x{decoded[0]:02x}')

coinbase = base64.b64decode('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=')
print(f'Coinbase length: {len(coinbase)} bytes') 
print(f'Coinbase hex: {coinbase.hex()}')



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from qbitcoin.core import config
from qbitcoin.core.Singleton import Singleton
from qbitcoin.core.miners.qryptonight7.CNv1PoWValidator import CNv1PoWValidator
from qbitcoin.core.miners.qrandomx.QRXPoWValidator import QRXPoWValidator


class PoWValidator(object, metaclass=Singleton):
    def __init__(self):
        self.qryptonight_7_pow_validator = CNv1PoWValidator()
        self.qryptonight_r_pow_validator = QRXPoWValidator()

    def verify_input(self, block_number, seed_height, seed_hash, mining_blob, target):
        if block_number < config.dev.hard_fork_heights[0]:
            return self.qryptonight_7_pow_validator.verify_input(mining_blob, target)
        else:
            return self.qryptonight_r_pow_validator.verify_input(block_number,
                                                                 seed_height,
                                                                 seed_hash,
                                                                 mining_blob,
                                                                 target)



from pyqrllib.pyqrllib import bin2hstr
from qbitcoin.core.FalconHelper import falcon_pk_to_address
from typing import Union

from qbitcoin.core.State import State
from qbitcoin.core.StateContainer import StateContainer
from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
from qbitcoin.core.MultiSigAddressState import MultiSigAddressState
from qbitcoin.core.misc import logger
from qbitcoin.core.txs.Transaction import Transaction


class TransferTransaction(Transaction):
    """
    SimpleTransaction for the transaction of QRL from one wallet to another.
    """

    def __init__(self, protobuf_transaction=None):
        super(TransferTransaction, self).__init__(protobuf_transaction)

    @property
    def addrs_to(self):
        return self._data.transfer.addrs_to

    @property
    def total_amount(self):
        total_amount = 0
        for amount in self.amounts:
            total_amount += amount
        return total_amount

    @property
    def amounts(self):
        return self._data.transfer.amounts

    @property
    def message_data(self):
        return self._data.transfer.message_data

    def get_data_bytes(self):
        tmptxhash = (self.master_addr +
                     self.fee.to_bytes(8, byteorder='big', signed=False) +
                     self.message_data)

        for index in range(0, len(self.addrs_to)):
            tmptxhash = (tmptxhash +
                         self.addrs_to[index] +
                         self.amounts[index].to_bytes(8, byteorder='big', signed=False))

        return tmptxhash

    @staticmethod
    def create(addrs_to: list, amounts: list, message_data: Union[bytes, None], fee: int, xmss_pk, master_addr: bytes = None):
        transaction = TransferTransaction()

        if master_addr:
            transaction._data.master_addr = master_addr

        transaction._data.public_key = bytes(xmss_pk)

        for addr_to in addrs_to:
            transaction._data.transfer.addrs_to.append(addr_to)

        for amount in amounts:
            transaction._data.transfer.amounts.append(amount)

        if message_data:
            transaction._data.transfer.message_data = message_data

        transaction._data.fee = int(fee)  # FIXME: Review conversions for quantities

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def _validate_custom(self):
        for amount in self.amounts:
            if amount == 0:
                logger.warning('Amount cannot be 0 - %s', self.amounts)
                logger.warning('Invalid TransferTransaction')
                return False

        if self.fee < 0:
            logger.info('TransferTransaction [%s] Invalid Fee = %d', bin2hstr(self.txhash), self.fee)
            return False

        if len(self.addrs_to) == 0:
            logger.warning("[TransferTransaction] No Addrs To found")
            return False

        if len(self.addrs_to) != len(self.amounts):
            logger.warning('[TransferTransaction] Mismatch number of addresses to & amounts')
            logger.warning('>> Length of addresses_to %s', len(self.addrs_to))
            logger.warning('>> Length of amounts %s', len(self.amounts))
            return False

        if not OptimizedAddressState.address_is_valid(self.addr_from):
            logger.warning('[TransferTransaction] Invalid address addr_from: %s', bin2hstr(self.addr_from))
            return False

        for addr_to in self.addrs_to:
            if not (OptimizedAddressState.address_is_valid(addr_to) or MultiSigAddressState.address_is_valid(addr_to)):
                logger.warning('[TransferTransaction] Invalid address addr_to: %s', bin2hstr(addr_to))
                return False

        return True

    # checks new tx validity based upon node statedb and node mempool.
    def _validate_extended(self, state_container: StateContainer):
        # Removed hard fork check for message data - always allow for development
        # Original check was:
        # if len(self.message_data) > 0:
        #     if state_container.block_number < state_container.current_dev_config.hard_fork_heights[0]:
        #         logger.warning("[TransferTransaction] Hard Fork Feature not yet activated")
        #         return False

        if len(self.addrs_to) > state_container.current_dev_config.transaction_multi_output_limit:
            logger.warning('[TransferTransaction] Number of addresses exceeds max limit')
            logger.warning('Number of addresses %s', len(self.addrs_to))
            logger.warning('Number of amounts %s', len(self.amounts))
            return False

        if len(self.message_data) > state_container.current_dev_config.message_max_length:
            logger.warning("[TransferTransaction] Message data is greater than message max length limit")
            logger.warning("Message data length %s", len(self.message_data))
            logger.warning("Message data length limit %s", state_container.current_dev_config.message_max_length)
            return False

        tx_balance = state_container.addresses_state[self.addr_from].balance
        total_amount = self.total_amount

        for addr_to in self.addrs_to:
            if MultiSigAddressState.address_is_valid(addr_to):
                if addr_to not in state_container.addresses_state:
                    logger.warning('[TransferTransaction] Multi Sig Address doesnt exist: %s', bin2hstr(addr_to))
                    return False

        if tx_balance < total_amount + self.fee:
            logger.info('State validation failed for %s because: Insufficient funds', bin2hstr(self.txhash))
            logger.info('balance: %s, fee: %s, amount: %s', tx_balance, self.fee, total_amount)
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)
        for addr_to in self.addrs_to:
            addresses_set.add(addr_to)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.total_amount + self.fee, subtract=True)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)

        addr_from_pk = falcon_pk_to_address(self.PK)
        for index in range(0, len(self.addrs_to)):
            addr_to = self.addrs_to[index]
            amount = self.amounts[index]
            address_state = state_container.addresses_state[addr_to]
            address_state.update_balance(state_container, amount)
            if addr_to in (self.addr_from, addr_from_pk):
                continue
            state_container.paginated_tx_hash.insert(address_state, self.txhash)

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container,
                                     self.total_amount + self.fee)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        addr_from_pk = falcon_pk_to_address(self.PK)
        for index in range(0, len(self.addrs_to)):
            addr_to = self.addrs_to[index]
            amount = self.amounts[index]
            address_state = state_container.addresses_state[addr_to]
            address_state.update_balance(state_container, amount, subtract=True)
            if addr_to in (self.addr_from, addr_from_pk):
                continue
            state_container.paginated_tx_hash.remove(address_state, self.txhash)

        return self._revert_state_changes_for_PK(state_container)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from qbitcoin.core import config
from qbitcoin.core.PaginatedData import PaginatedData
from qbitcoin.core.misc import logger
from qbitcoin.core.PaginatedFalconPK import PaginatedFalconPK
from qbitcoin.generated import qbit_pb2


class PaginatedBitfield(PaginatedData):
    """
    This class maintains backward compatibility with old OTS tracking,
    but delegates to the new PaginatedFalconPK for actual functionality.
    It's kept to minimize changes in the codebase during migration.
    """
    def __init__(self, write_access: bool, db):
        super(PaginatedBitfield, self).__init__(b'bitfield', write_access, db)
        # Create a new instance of PaginatedFalconPK for delegating calls
        self._falcon_pk_tracker = PaginatedFalconPK(write_access, db)

    def generate_bitfield_key(self, address, page):
        # Kept for backward compatibility
        return self.name + b'_' + address + b'_' + page.to_bytes(8, byteorder='big', signed=False)

    def load_bitfield_and_ots_key_reuse(self, address, ots_key_index) -> bool:
        # This is deprecated - we don't use OTS with Falcon
        logger.warning("load_bitfield_and_ots_key_reuse is deprecated - Falcon doesn't use OTS keys")
        return False

    @staticmethod
    def ots_key_reuse(ots_bitfield, ots_key_index) -> bool:
        # Deprecated - return False to avoid key reuse errors
        logger.warning("ots_key_reuse is deprecated - Falcon doesn't use OTS keys")
        return False

    def set_ots_key(self, addresses_state: dict, address, public_key):
        """
        Transition method - instead of tracking OTS keys, track Falcon public keys
        """
        # Simply delegate to the falcon public key tracker
        self._falcon_pk_tracker.add_used_key(addresses_state, address, public_key)

    def update_used_page_in_address_state(self, address, addresses_state: dict, page: int):
        # This no longer needs to be implemented for Falcon
        pass

    def unset_ots_key(self, addresses_state: dict, address, public_key):
        """
        Transition method - removes a public key from tracking
        Used during rollbacks
        """
        self._falcon_pk_tracker.remove_used_key(addresses_state, address, public_key)

    def load_bitfield(self, address, ots_key_index):
        # This is deprecated - kept for backward compatibility
        pass

    def get_paginated_data(self, key, page):
        """
        Maintained for backward compatibility
        """
        try:
            pbData = self.db.get_raw(self.name + b'_' + key + b'_' + page.to_bytes(8, byteorder='big', signed=False))
            data_list = qbit_pb2.DataList()
            data_list.ParseFromString(bytes(pbData))
            return list(data_list.values)
        except KeyError:
            return []
        except Exception as e:
            logger.error('[get_paginated_data] Exception for %s', self.name)
            logger.exception(e)
            raise

    def put_addresses_bitfield(self, batch):
        """
        Save all data to db
        """
        if not self.write_access:
            return
            
        # Save falcon public keys
        self._falcon_pk_tracker.put_used_keys(batch)
        
        # Also save traditional bitfields for backward compatibility
        for key in self.key_value:
            data_list = qbit_pb2.DataList(values=self.key_value[key])
            self.db.put_raw(key,
                            data_list.SerializeToString(),
                            batch)



"""
Falcon Helper functions for working with addresses and public keys
"""
import hashlib


def falcon_pk_to_address(public_key: bytes) -> bytes:
    """
    Convert a Falcon public key to a QRL address in bytes format
    
    Args:
        public_key (bytes): Falcon public key
        
    Returns:
        bytes: QRL address in bytes format (25 bytes total)
    """
    # Perform SHA-256 hash
    sha256_hash = hashlib.sha256(public_key).digest()
    
    # Perform RIPEMD-160 hash on the SHA-256 hash
    ripemd160_hash = hashlib.new('ripemd160')
    ripemd160_hash.update(sha256_hash)
    hash160 = ripemd160_hash.digest()
    
    # Create binary address with proper format:
    # [0] - Address type (0x01 for standard Qbitcoin address with Falcon)
    # [1:21] - RIPEMD160 hash of the public key (20 bytes)
    address_bytes = b'\x01' + hash160
    
    # Add a 4-byte checksum (double SHA-256 of the first 21 bytes)
    checksum = hashlib.sha256(hashlib.sha256(address_bytes).digest()).digest()[:4]
    binary_address = address_bytes + checksum
    
    return binary_address


def is_falcon_public_key(pk: bytes) -> bool:
    """
    Checks if the provided public key is a valid Falcon public key
    
    Args:
        pk (bytes): Public key to check
        
    Returns:
        bool: True if the public key is likely a Falcon public key
    """
    from pqcrypto.sign.falcon_512 import PUBLIC_KEY_SIZE
    
    # Falcon-512 has a specific public key size
    return len(pk) == PUBLIC_KEY_SIZE



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import decimal
from decimal import Decimal
from qbitcoin.core.config import DevConfig


def get_halving_interval(dev_config: DevConfig) -> int:
    """
    Return the number of blocks between halvings (3.805 years at 1 minute per block).
    3.805 years * 365.25 days/year * 24 hours/day * 60 minutes/hour = ~2,000,000 blocks
    """
    return 2000000  # ~3.805 years

def get_initial_block_reward(dev_config: DevConfig) -> Decimal:
    """
    Return the initial block reward: 2.5 Qbitcoin = 2.5 * 10^9 Quark
    """
    return Decimal('2.5') * dev_config.quark_per_qbitcoin


def remaining_emission(block_n, dev_config: DevConfig) -> Decimal:
    """
    Calculate remaining emission at block_n using halving mechanism.
    Starting with 10M QRL available for mining (30M max - 20M genesis)
    Initial reward: 2.5 Qbitcoin per block
    Halving every 2 years (1,051,200 blocks)
    """
    if block_n <= 0:
        return dev_config.coin_remaining_at_genesis * dev_config.quark_per_qbitcoin
    
    halving_interval = get_halving_interval(dev_config)
    initial_reward = get_initial_block_reward(dev_config)
    
    total_mined = Decimal('0')
    current_reward = initial_reward
    blocks_processed = 0
    
    while blocks_processed < block_n:
        # How many blocks in current halving period?
        blocks_in_period = min(halving_interval, block_n - blocks_processed)
        
        # Add rewards for this period
        total_mined += current_reward * blocks_in_period
        
        blocks_processed += blocks_in_period
        current_reward = current_reward / 2  # Halve the reward
    
    remaining = dev_config.coin_remaining_at_genesis * dev_config.quark_per_qbitcoin - total_mined
    return remaining.quantize(Decimal('1.'), rounding=decimal.ROUND_DOWN)


def block_reward(block_number: int, dev_config: DevConfig) -> Decimal:
    """
    Calculate block reward using halving mechanism.
    Initial reward: 2.5 Qbitcoin per block
    Halving every 2 years (1,051,200 blocks)
    """
    if block_number <= 0:
        return Decimal('0')
    
    halving_interval = get_halving_interval(dev_config)
    initial_reward = get_initial_block_reward(dev_config)
    
    # Calculate which halving period we're in (0-based)
    halving_period = (block_number - 1) // halving_interval
    
    # Calculate current reward (halved for each period)
    current_reward = initial_reward / (Decimal('2') ** halving_period)
    
    return current_reward.quantize(Decimal('1.'), rounding=decimal.ROUND_DOWN)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

import os
import json
import time
import threading
import math
from typing import Dict, List, Optional, Set

from pyqrllib.pyqrllib import bin2hstr, hstr2bin

from qbitcoin.core.config import DevConfig, UserConfig
from qbitcoin.core.misc import logger
from qbitcoin.core.txs.TransferTransaction import TransferTransaction
from qbitcoin.core.Singleton import Singleton
from qbitcoin.core.DonorWalletManager import DonorWalletManager


class StakeInfo:
    """Information about a staker"""
    
    def __init__(self, address: bytes, balance: int, peer_ip: str = None, is_local: bool = False):
        self.address = address
        self.balance = balance
        self.stake_start_time = time.time()
        self.last_seen_time = time.time()
        self.is_online = True
        self.peer_ip = peer_ip  # IP of peer that shared this staker info
        self.is_local = is_local  # True if this is our own node's staker
        self.last_balance_check = time.time()
        self.balance_verified = False
        
    def update_last_seen(self):
        """Update last seen time to current time"""
        self.last_seen_time = time.time()
        self.is_online = True
        
    def check_online_status(self, max_offline_time: int) -> bool:
        """Check if staker is still online based on last seen time"""
        current_time = time.time()
        offline_time = current_time - self.last_seen_time
        self.is_online = offline_time <= max_offline_time
        return self.is_online
        
    def get_stake_share(self, total_stake: int) -> float:
        """Calculate this staker's share of total stake"""
        if total_stake == 0:
            return 0.0
        return self.balance / total_stake
        
    def serialize(self) -> dict:
        """Serialize stake info to dictionary"""
        return {
            'address': bin2hstr(self.address),
            'balance': self.balance,
            'stake_start_time': self.stake_start_time,
            'last_seen_time': self.last_seen_time,
            'is_online': self.is_online,
            'peer_ip': self.peer_ip,
            'is_local': self.is_local,
            'last_balance_check': self.last_balance_check,
            'balance_verified': self.balance_verified
        }
        
    @classmethod
    def deserialize(cls, data: dict) -> 'StakeInfo':
        """Deserialize stake info from dictionary"""
        stake_info = cls(
            address=hstr2bin(data['address']),
            balance=data['balance'],
            peer_ip=data.get('peer_ip'),
            is_local=data.get('is_local', False)
        )
        stake_info.stake_start_time = data.get('stake_start_time', time.time())
        stake_info.last_seen_time = data.get('last_seen_time', time.time())
        stake_info.is_online = data.get('is_online', True)
        stake_info.last_balance_check = data.get('last_balance_check', time.time())
        stake_info.balance_verified = data.get('balance_verified', False)
        return stake_info


class StakingManager(object, metaclass=Singleton):
    """
    Simplified staking mechanism for Qbitcoin
    Only stores staker addresses and verifies balances from blockchain
    """
    
    def __init__(self):
        self.stakers: Set[bytes] = set()  # Only store addresses
        self.last_reward_block = 0
        self.lock = threading.RLock()
        self.chain_manager = None
        self.p2p_factory = None
        self.dev_config = None
        
        # For P2P staker synchronization (like blocks)
        self.pending_broadcasts = []  # Keep for compatibility
        self.is_node_synced = False
        
        # Initialize persistent storage
        user_config = UserConfig.getInstance()
        self.stakers_db_path = os.path.join(user_config.qrl_dir, 'data', 'stakers.json')
        self.ensure_data_directory()
        
        # Initialize donor wallet manager
        self.donor_wallet_manager = DonorWalletManager()
        
        # Load existing stakers from database
        self.load_stakers_from_db()
        
        # Auto-load donor wallet if available
        wallet_loaded = self.auto_load_donor_wallet()
        if wallet_loaded:
            logger.info("✓ DONOR WALLET LOADED SUCCESSFULLY - Rewards will be distributed automatically")
        else:
            logger.debug("No donor wallet found - rewards will not be distributed")
        
        # Start reward monitoring thread
        self._reward_monitoring = True
        self._reward_thread = threading.Thread(target=self._monitor_rewards, daemon=True)
        self._reward_thread.start()
        
    def ensure_data_directory(self):
        """Ensure data directory exists"""
        data_dir = os.path.dirname(self.stakers_db_path)
        os.makedirs(data_dir, exist_ok=True)
        
    def load_stakers_from_db(self):
        """Load stakers from persistent storage - only addresses"""
        try:
            if os.path.exists(self.stakers_db_path):
                with open(self.stakers_db_path, 'r') as f:
                    data = json.load(f)
                
                # Handle both old and new database formats
                stakers_data = data.get('stakers', [])
                loaded_count = 0  # Initialize counter
                
                for item in stakers_data:
                    try:
                        # Handle new format (list of address strings)
                        if isinstance(item, str):
                            # Validate Qbitcoin address format (50 hex chars for 25 bytes)
                            if len(item) == 50:  # Qbitcoin addresses are 50 hex chars (25 bytes)
                                try:
                                    address_data = hstr2bin(item)
                                    # Convert tuple to bytes if needed (hstr2bin returns tuple in this version)
                                    if isinstance(address_data, tuple):
                                        address_bytes = bytes(address_data)
                                    else:
                                        address_bytes = address_data
                                        
                                    self.stakers.add(address_bytes)
                                    loaded_count += 1
                                    logger.debug("Loaded staker address: %s", item)
                                except Exception as e:
                                    logger.warning("Failed to convert address %s: %s", item, str(e))
                            else:
                                logger.warning("Invalid address format: %s (length: %d)", item, len(item))
                        
                        # Handle old format (dict with address field)
                        elif isinstance(item, dict) and 'address' in item:
                            address_hex = item['address']
                            if isinstance(address_hex, str) and len(address_hex) == 50:
                                try:
                                    address_data = hstr2bin(address_hex)
                                    # Convert tuple to bytes if needed (hstr2bin returns tuple in this version)
                                    if isinstance(address_data, tuple):
                                        address_bytes = bytes(address_data)
                                    else:
                                        address_bytes = address_data
                                        
                                    self.stakers.add(address_bytes)
                                    loaded_count += 1
                                    logger.debug("Loaded staker address from dict: %s", address_hex)
                                except Exception as e:
                                    logger.warning("Failed to convert address %s: %s", address_hex, str(e))
                            else:
                                logger.warning("Invalid address in old format: %s", address_hex)
                                
                        # Handle tuple format (raw bytes as tuple) - this is valid for Qbitcoin
                        elif isinstance(item, (list, tuple)) and len(item) == 25:
                            if all(isinstance(x, int) and 0 <= x <= 255 for x in item):
                                address_bytes = bytes(item)
                                from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
                                if OptimizedAddressState.address_is_valid(address_bytes):
                                    self.stakers.add(address_bytes)
                                    loaded_count += 1
                                    logger.debug("Loaded tuple address: %s", bin2hstr(address_bytes)[:10] + "...")
                                else:
                                    logger.warning("Invalid tuple address format: %s", item)
                            else:
                                logger.warning("Invalid tuple data: %s", item)
                        else:
                            logger.warning("Unknown staker data format: %s", type(item))
                            
                    except Exception as e:
                        logger.warning("Failed to load staker address %s: %s", str(item)[:20], str(e))
                        
                self.last_reward_block = data.get('last_reward_block', 0)
                logger.info("Loaded %d staker addresses from database", len(self.stakers))
                    
            else:
                logger.info("No existing stakers database found")
        except Exception as e:
            logger.error("Failed to load stakers database: %s", str(e))
            # Reset stakers to empty set if loading fails
            self.stakers = set()
            
    def save_stakers_to_db(self):
        """Save stakers to persistent storage - only addresses"""
        try:
            data = {
                'stakers': [bin2hstr(address) for address in self.stakers],
                'last_reward_block': self.last_reward_block,
                'last_saved': time.time()
            }
            
            # Write to temporary file first, then rename for atomicity
            temp_path = self.stakers_db_path + '.tmp'
            with open(temp_path, 'w') as f:
                json.dump(data, f, indent=2)
                
            os.rename(temp_path, self.stakers_db_path)
            logger.debug("Saved %d staker addresses to database", len(self.stakers))
        except Exception as e:
            logger.error("Failed to save stakers database: %s", str(e))
        
    def set_chain_manager(self, chain_manager):
        """Set the chain manager instance"""
        self.chain_manager = chain_manager
        
    def verify_staker_balance(self, address: bytes) -> Optional[int]:
        """
        Verify staker balance from blockchain
        
        Args:
            address: Staker's address
            
        Returns:
            Balance if valid, None if invalid
        """
        if not self.chain_manager:
            logger.error("Chain manager not available for balance verification")
            return None
            
        try:
            # Ensure address is bytes type
            if isinstance(address, str):
                if address.startswith('Q'):
                    address_data = hstr2bin(address[1:])  # Remove Q prefix
                else:
                    address_data = hstr2bin(address)
                
                # Convert tuple to bytes if needed (hstr2bin returns tuple in this version)
                if isinstance(address_data, tuple):
                    address = bytes(address_data)
                else:
                    address = address_data
                    
            elif isinstance(address, tuple):
                # Convert tuple of integers to bytes (this is valid for Qbitcoin addresses)
                if len(address) == 25 and all(isinstance(x, int) and 0 <= x <= 255 for x in address):
                    address = bytes(address)
                    logger.debug("Converted tuple address to bytes: %s", bin2hstr(address)[:10] + "...")
                else:
                    logger.error("Invalid tuple address format: %s", address)
                    return None
            elif not isinstance(address, bytes):
                logger.error("Invalid address type: %s", type(address))
                return None
                
            # Validate address format using existing validation
            from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
            if not OptimizedAddressState.address_is_valid(address):
                logger.warning("Address format is invalid: %s", bin2hstr(address))
                return None
                
            address_state = self.chain_manager.get_optimized_address_state(address)
            if not address_state:
                logger.warning("Address %s not found in blockchain", bin2hstr(address))
                return None
                
            balance = address_state.balance
            
            # Check positive balance
            if balance <= 0:
                logger.warning("Address %s has zero or negative balance: %d", bin2hstr(address), balance)
                return None
                
            # Check minimum balance requirement
            if self.dev_config and balance < self.dev_config.staking_min_balance:
                logger.warning("Address %s balance below minimum: %d < %d", 
                             bin2hstr(address), balance, self.dev_config.staking_min_balance)
                return None
                
            return balance
            
        except Exception as e:
            logger.error("Error verifying balance for %s: %s", bin2hstr(address), str(e))
            return None
        
    def set_p2p_factory(self, p2p_factory):
        """Set the P2P factory"""
        self.p2p_factory = p2p_factory
        
    def set_grpc_client(self, grpc_client):
        """Set the gRPC client for transaction broadcasting"""
        self.grpc_client = grpc_client
        
    def set_dev_config(self, dev_config: DevConfig):
        """Set the development configuration"""
        self.dev_config = dev_config
        
    def validate_and_cleanup_stakers(self) -> int:
        """
        Validate all current stakers and remove those with insufficient balance
        
        Returns:
            int: Number of stakers removed
        """
        if not self.chain_manager or not self.dev_config:
            logger.warning("Cannot validate stakers: chain_manager or dev_config not available")
            return 0
            
        removed_count = 0
        stakers_to_remove = []
        
        with self.lock:
            logger.info("Validating %d stakers for minimum balance requirement", len(self.stakers))
            
            for address in self.stakers.copy():  # Use copy to avoid modification during iteration
                try:
                    # Verify current balance
                    balance = self.verify_staker_balance(address)
                    
                    if balance is None:
                        # Balance verification failed (address not found, below minimum, etc.)
                        stakers_to_remove.append(address)
                        logger.info("Removing staker %s: balance verification failed", bin2hstr(address))
                    else:
                        logger.debug("Staker %s validated: balance %d QBC", bin2hstr(address), balance // 10**9)
                        
                except Exception as e:
                    logger.error("Error validating staker %s: %s", bin2hstr(address), str(e))
                    stakers_to_remove.append(address)
            
            # Remove invalid stakers
            for address in stakers_to_remove:
                if address in self.stakers:
                    self.stakers.discard(address)
                    removed_count += 1
                    logger.info("Removed staker %s due to insufficient balance", bin2hstr(address))
                    
                    # Broadcast removal to peers
                    self.broadcast_staker_to_peers(address, 'remove')
            
            # Save updated staker list if any were removed
            if removed_count > 0:
                self.save_stakers_to_db()
                logger.info("Staker cleanup completed: removed %d invalid stakers, %d remaining", 
                           removed_count, len(self.stakers))
            else:
                logger.debug("Staker validation completed: all %d stakers are valid", len(self.stakers))
        
        return removed_count
    
    def periodic_staker_validation(self):
        """
        Periodically validate stakers (called from monitoring thread)
        """
        try:
            # Only run validation if node is synced and we have stakers
            if not self.is_node_synced or not self.stakers:
                return
                
            # Run validation every 10 minutes (600 seconds)
            current_time = time.time()
            if not hasattr(self, '_last_validation_time'):
                self._last_validation_time = 0
                
            if current_time - self._last_validation_time >= 600:  # 10 minutes
                logger.debug("Starting periodic staker validation...")
                removed_count = self.validate_and_cleanup_stakers()
                
                if removed_count > 0:
                    logger.info("Periodic staker validation: removed %d invalid stakers", removed_count)
                
                self._last_validation_time = current_time
                
        except Exception as e:
            logger.error("Error in periodic staker validation: %s", str(e))
    
    def validate_staker_after_transaction(self, address: bytes):
        """
        Validate a specific staker after a transaction that might affect their balance
        
        Args:
            address: Address to validate
        """
        try:
            if not self.chain_manager or not self.dev_config:
                return
                
            # Check if this address is a staker
            if address not in self.stakers:
                return
                
            # Verify current balance
            balance = self.verify_staker_balance(address)
            
            if balance is None:
                # Remove staker due to insufficient balance
                with self.lock:
                    if address in self.stakers:
                        self.stakers.discard(address)
                        self.save_stakers_to_db()
                        self.broadcast_staker_to_peers(address, 'remove')
                        
                        balance_qbc = 0
                        try:
                            address_state = self.chain_manager.get_optimized_address_state(address)
                            if address_state:
                                balance_qbc = address_state.balance / 10**9
                        except:
                            pass
                            
                        logger.info("Staker %s removed after transaction: balance %.3f QBC below minimum %.0f QBC", 
                                   bin2hstr(address), balance_qbc, self.dev_config.staking_min_balance / 10**9)
            else:
                logger.debug("Staker %s still valid after transaction: balance %d quarks", 
                           bin2hstr(address), balance)
                
        except Exception as e:
            logger.error("Error validating staker %s after transaction: %s", bin2hstr(address), str(e))
    
    def force_validate_all_stakers(self) -> Dict[str, int]:
        """
        Force immediate validation of all stakers (for testing/debugging)
        
        Returns:
            dict: Statistics about the validation
        """
        try:
            logger.info("Force validating all stakers...")
            
            initial_count = len(self.stakers)
            removed_count = self.validate_and_cleanup_stakers()
            remaining_count = len(self.stakers)
            
            stats = {
                'initial_stakers': initial_count,
                'removed_stakers': removed_count,
                'remaining_stakers': remaining_count
            }
            
            logger.info("Force validation completed: %d initial, %d removed, %d remaining", 
                       initial_count, removed_count, remaining_count)
            
            return stats
            
        except Exception as e:
            logger.error("Error in force validation: %s", str(e))
            return {'error': str(e)}
        
    def load_donor_wallet(self, password: str) -> bool:
        """
        Load donor wallet with password
        
        Args:
            password: Password to decrypt donor wallet
            
        Returns:
            bool: True if wallet was loaded successfully
        """
        return self.donor_wallet_manager.load_donor_wallet(password) is not None
        
    def auto_load_donor_wallet(self) -> bool:
        """
        Automatically load donor wallet if available
        
        Returns:
            bool: True if wallet was loaded successfully
        """
        try:
            
            from qbitcoin.core.config import UserConfig
            user_config = UserConfig.getInstance()
            donor_wallets_dir = os.path.join(user_config.qrl_dir, 'donor_wallets')
            
            # First try to load from donor_wallets directory
            if os.path.exists(donor_wallets_dir):
                logger.info("Checking donor wallets directory: %s", donor_wallets_dir)
                # Find unencrypted donor wallet files
                for filename in os.listdir(donor_wallets_dir):
                    if filename.endswith('_donor.json'):
                        wallet_path = os.path.join(donor_wallets_dir, filename)
                        logger.info("Found donor wallet file: %s", filename)
                        try:
                            with open(wallet_path, 'r') as f:
                                wallet_data = json.load(f)
                                
                            # Check if it's unencrypted
                            if not wallet_data.get('encrypted', True):  # Default to encrypted if not specified
                                logger.info("Loading unencrypted donor wallet: %s", filename)
                                
                                # Load the wallet directly without password
                                if self.donor_wallet_manager.load_unencrypted_donor_wallet(wallet_data):
                                    logger.info("✓ Successfully auto-loaded donor wallet from %s", filename)
                                    logger.info("✓ Donor wallet address: %s", wallet_data.get('address', 'unknown'))
                                    return True
                                else:
                                    logger.warning("✗ Failed to load unencrypted donor wallet: %s", filename)
                            else:
                                logger.info("Skipping encrypted donor wallet: %s", filename)
                        except Exception as e:
                            logger.warning("Error reading donor wallet %s: %s", filename, str(e))
            else:
                logger.info("Donor wallets directory not found: %s", donor_wallets_dir)
            
            # REMOVED: No longer automatically load genesis wallet as donor wallet
            # This was causing confusion - donor wallets should be explicitly created
            logger.debug("No donor wallet found - rewards will not be distributed")
            return False
            
        except Exception as e:
            logger.error("Error in auto_load_donor_wallet: %s", str(e))
            return False
        
    def is_donor_wallet_loaded(self) -> bool:
        """Check if donor wallet is currently loaded and files still exist"""
        if not self.donor_wallet_manager.is_wallet_loaded():
            return False
            
        # Verify that donor wallet files still exist
        from qbitcoin.core.config import UserConfig
        user_config = UserConfig.getInstance()
        donor_wallets_dir = os.path.join(user_config.qrl_dir, 'donor_wallets')
        
        if not os.path.exists(donor_wallets_dir):
            # Directory doesn't exist, unload the wallet
            logger.debug("Donor wallets directory not found, unloading wallet from memory")
            self.donor_wallet_manager.unload_wallet()
            return False
            
        # Check if any donor wallet files exist
        donor_files = [f for f in os.listdir(donor_wallets_dir) if f.endswith('_donor.json')]
        if not donor_files:
            # No donor wallet files exist, unload the wallet
            logger.debug("No donor wallet files found, unloading wallet from memory")
            self.donor_wallet_manager.unload_wallet()
            return False
            
        return True
        
    def get_donor_wallet_address(self) -> Optional[str]:
        """Get donor wallet address"""
        wallet_info = self.donor_wallet_manager.get_wallet_info()
        if wallet_info:
            return wallet_info['address']
        return None
            
    def add_staker(self, address: bytes, peer_ip: str = None) -> bool:
        """
        Add a new staker to the staking pool
        
        Args:
            address: Staker's address
            peer_ip: IP of peer that shared this staker (None for local stakers)
            
        Returns:
            bool: True if staker was added successfully
        """
        if not self.dev_config or not self.dev_config.staking_enabled:
            logger.warning("Staking is not enabled")
            return False
            
        with self.lock:
            # Ensure address is bytes type and handle various formats
            if isinstance(address, str):
                try:
                    if address.startswith('Q'):
                        address_data = hstr2bin(address[1:])  # Remove Q prefix
                    else:
                        address_data = hstr2bin(address)
                    
                    # Convert tuple to bytes if needed (hstr2bin returns tuple in this version)
                    if isinstance(address_data, tuple):
                        address = bytes(address_data)
                    else:
                        address = address_data
                        
                except Exception as e:
                    logger.error("Invalid address format: %s - %s", address, str(e))
                    return False
            elif isinstance(address, tuple):
                # Convert tuple of integers to bytes (this is valid for Qbitcoin addresses)
                if len(address) == 25 and all(isinstance(x, int) and 0 <= x <= 255 for x in address):
                    address = bytes(address)
                    logger.debug("Converted tuple address to bytes: %s", bin2hstr(address)[:10] + "...")
                else:
                    logger.error("Invalid tuple address format: %s", address)
                    return False
            elif not isinstance(address, bytes):
                logger.error("Invalid address type: %s", type(address))
                return False
                
            # Validate address format using existing validation
            from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
            if not OptimizedAddressState.address_is_valid(address):
                logger.warning("Address format is invalid: %s", bin2hstr(address))
                return False
            
            # Check if address is already in the staking list (no duplicates)
            if address in self.stakers:
                logger.info("Address %s is already staking", bin2hstr(address))
                return False
                
            # Verify balance from blockchain
            balance = self.verify_staker_balance(address)
            if balance is None:
                logger.error("Failed to verify balance for address %s", bin2hstr(address))
                return False
                
            # Add staker address to set
            self.stakers.add(address)
            
            # Save to database
            self.save_stakers_to_db()
            
            # Broadcast to peers (only if this is a local staker, not from peer)
            if not peer_ip:
                self.broadcast_staker_to_peers(address, 'add')
            
            logger.info("Staker added: %s with verified balance %d", bin2hstr(address), balance)
            return True
            
    def remove_staker(self, address: bytes, reason: str = "manual") -> bool:
        """
        Remove a staker from the staking pool
        
        Args:
            address: Staker's address
            reason: Reason for removal
            
        Returns:
            bool: True if staker was removed
        """
        with self.lock:
            if address not in self.stakers:
                return False
                
            # Remove from set
            self.stakers.discard(address)
            
            # Save to database
            self.save_stakers_to_db()
            
            # Broadcast removal to peers (only if not removed by peer)
            if not reason.startswith("removed by peer"):
                self.broadcast_staker_to_peers(address, 'remove')
            
            logger.info("Staker removed: %s (reason: %s)", bin2hstr(address), reason)
            return True
            
    def get_active_stakers(self) -> Dict[bytes, int]:
        """Get list of active stakers with verified balances from blockchain"""
        with self.lock:
            active_stakers = {}
            invalid_stakers = []
            
            for address in self.stakers:
                balance = self.verify_staker_balance(address)
                if balance is not None:
                    active_stakers[address] = balance
                else:
                    # Mark for removal if balance verification fails
                    invalid_stakers.append(address)
                    
            # Remove invalid stakers
            for address in invalid_stakers:
                logger.info("Removing invalid staker: %s", bin2hstr(address))
                self.stakers.discard(address)
                
            # Save changes if any stakers were removed
            if invalid_stakers:
                self.save_stakers_to_db()
                
            return active_stakers
            
    def get_total_stake(self) -> int:
        """Get total staking amount from active stakers"""
        active_stakers = self.get_active_stakers()
        return sum(active_stakers.values())
            
    def should_distribute_rewards(self, current_block_number: int) -> bool:
        """Check if it's time to distribute rewards"""
        if not self.dev_config:
            logger.debug("No dev_config available for reward distribution")
            return False
            
        if not self.is_donor_wallet_loaded():
            logger.debug("Donor wallet not loaded for reward distribution")
            return False
            
        blocks_since_last_reward = current_block_number - self.last_reward_block
        should_distribute = blocks_since_last_reward >= self.dev_config.staking_reward_interval
        
        logger.debug("Reward check: current_block=%d, last_reward_block=%d, interval=%d, should_distribute=%s",
                    current_block_number, self.last_reward_block, 
                    self.dev_config.staking_reward_interval, should_distribute)
        
        return should_distribute
        
    def distribute_rewards(self, current_block_number: int) -> List[dict]:
        """
        Distribute staking rewards to active stakers based on their proportional contribution
        
        Args:
            current_block_number: Current block number
            
        Returns:
            List of reward transaction details
        """
        if not self.should_distribute_rewards(current_block_number):
            return []
            
        with self.lock:
            if not self.stakers:
                logger.info("No stakers registered for reward distribution")
                return []
                
            # Verify each staker's balance and collect valid stakers
            valid_stakers = {}  # address -> verified_balance
            total_verified_stake = 0
            
            logger.info("Verifying balances for %d stakers before reward distribution", len(self.stakers))
            
            for address in self.stakers:
                # Verify balance from blockchain
                verified_balance = self.verify_staker_balance(address)
                
                if verified_balance is None:
                    logger.warning("Could not verify balance for staker %s, excluding from rewards", bin2hstr(address))
                    continue
                    
                # Check if balance meets minimum requirements
                if verified_balance <= 0:
                    logger.warning("Staker %s has zero balance, excluding from rewards", bin2hstr(address))
                    continue
                    
                if verified_balance < self.dev_config.staking_min_balance:
                    logger.warning("Staker %s balance below minimum (%d < %d), excluding from rewards", 
                                 bin2hstr(address), verified_balance, self.dev_config.staking_min_balance)
                    continue
                
                # Valid staker - add to distribution list
                valid_stakers[address] = verified_balance
                total_verified_stake += verified_balance
                
            if not valid_stakers:
                logger.info("No valid stakers found for reward distribution")
                return []
                
            if total_verified_stake == 0:
                logger.warning("Total verified stake is zero, cannot distribute rewards")
                return []
                
            total_reward = self.dev_config.staking_reward_amount
            logger.info("Distributing %d QBC total rewards to %d valid stakers (total stake: %d QBC)", 
                       total_reward // 1000000000, len(valid_stakers), total_verified_stake // 1000000000)
            
            # Calculate proportional rewards for each staker
            reward_outputs = []
            total_distributed = 0
            
            for address, verified_balance in valid_stakers.items():
                # Calculate this staker's share of total stake
                stake_percentage = verified_balance / total_verified_stake
                reward_amount = int(total_reward * stake_percentage)
                
                if reward_amount > 0:
                    reward_outputs.append({
                        'recipient': address,
                        'amount': reward_amount,
                        'stake_balance': verified_balance,
                        'stake_percentage': stake_percentage * 100,
                        'block_number': current_block_number
                    })
                    total_distributed += reward_amount
                    
                    logger.debug("Staker %s: %d QBC stake (%.2f%%) = %d QBC reward", 
                               bin2hstr(address), verified_balance // 1000000000, 
                               stake_percentage * 100, reward_amount // 1000000000)
            
            if not reward_outputs:
                logger.warning("No reward outputs generated")
                return []
                
            logger.info("Generated %d reward outputs, total distributed: %d QBC", 
                       len(reward_outputs), total_distributed // 1000000000)
            
            # Create reward transactions if donor wallet is loaded
            if self.is_donor_wallet_loaded() and self.chain_manager:
                try:
                    self._create_reward_transactions(reward_outputs)
                    self.last_reward_block = current_block_number
                    self.save_stakers_to_db()  # Save updated last reward block
                    logger.info("Reward distribution completed for block %d", current_block_number)
                except Exception as e:
                    logger.error("Failed to create reward transactions: %s", str(e))
            else:
                logger.warning("Cannot create reward transactions - donor wallet not loaded or chain manager not available")
            
            return reward_outputs
            
    def _create_reward_transactions(self, reward_outputs: List[dict]) -> List[dict]:
        """
        Create reward transactions with multiple outputs
        Split into multiple transactions if outputs exceed maximum
        
        Args:
            reward_outputs: List of reward output dictionaries
            
        Returns:
            List of created transaction hashes
        """
        if not self.is_donor_wallet_loaded():
            logger.warning("Donor wallet not loaded for reward transactions")
            return []
            
        max_outputs = getattr(self.dev_config, 'staking_max_outputs_per_tx', 100)
        transactions_created = []
        
        # Split rewards into chunks based on max outputs per transaction
        for i in range(0, len(reward_outputs), max_outputs):
            chunk = reward_outputs[i:i + max_outputs]
            
            try:
                tx_hash = self._create_single_reward_transaction(chunk)
                if tx_hash:
                    transactions_created.append({
                        'transaction_hash': tx_hash,
                        'outputs_count': len(chunk),
                        'total_amount': sum(output['amount'] for output in chunk)
                    })
                    logger.info("Created reward transaction %s with %d outputs", 
                              tx_hash, len(chunk))
                else:
                    logger.error("Failed to create reward transaction for chunk %d", i // max_outputs + 1)
                    
            except Exception as e:
                logger.error("Error creating reward transaction chunk %d: %s", 
                           i // max_outputs + 1, str(e))
                
        return transactions_created
        
    def _create_single_reward_transaction(self, reward_outputs: List[dict]) -> Optional[str]:
        """
        Create a single reward transaction with multiple outputs
        
        Args:
            reward_outputs: List of reward outputs for this transaction
            
        Returns:
            Transaction hash if successful, None otherwise
        """
        try:
            donor_address = self.donor_wallet_manager.get_address_bytes()
            donor_private_key = self.donor_wallet_manager.get_private_key_bytes()
            
            if not donor_address or not donor_private_key:
                logger.error("Cannot get donor wallet keys")
                return None
                
            # Prepare transaction outputs
            addrs_to = []
            amounts = []
            
            for output in reward_outputs:
                addrs_to.append(output['recipient'])
                amounts.append(output['amount'])
                
            # Calculate total amount and fee
            total_amount = sum(amounts)
            fee = 1000000  # 0.001 QBC fee in Quarks
            
            # Create transfer transaction using the same pattern as create_transaction.py
            logger.info("Creating reward transaction: %d QBC from %s to %d recipients", 
                       total_amount // 1000000000, 
                       bin2hstr(donor_address), 
                       len(addrs_to))
            
            # Get donor wallet info
            donor_wallet_info = self.donor_wallet_manager.get_wallet_info()
            if not donor_wallet_info:
                logger.error("Cannot get donor wallet info")
                return None
                
            # Convert public key from hex to bytes
            if 'public_key_hex' in donor_wallet_info:
                public_key_bytes = bytes.fromhex(donor_wallet_info['public_key_hex'])
            elif 'public_key' in donor_wallet_info:
                public_key_bytes = bytes.fromhex(donor_wallet_info['public_key'])
            else:
                logger.error("No public key found in donor wallet info")
                return None
                
            logger.debug("Public key type: %s, length: %d", type(public_key_bytes), len(public_key_bytes))
                
            # Create transaction manually (like create_transaction.py does)
            tx = TransferTransaction()
            tx._data.public_key = public_key_bytes
            
            # Add all recipient addresses and amounts
            for i, output in enumerate(reward_outputs):
                recipient_bytes = output['recipient']
                # Ensure recipient is bytes (convert if needed)
                if not isinstance(recipient_bytes, bytes):
                    recipient_bytes = bytes(recipient_bytes)
                    
                logger.debug("Adding recipient %d: type=%s, length=%d", i, type(recipient_bytes), len(recipient_bytes))
                tx._data.transfer.addrs_to.append(recipient_bytes)
                tx._data.transfer.amounts.append(output['amount'])
            
            # Set fee (same as create_transaction.py)
            tx._data.fee = 1000000  # 1 million quark fee
            
            # Set master_addr to donor address (like create_transaction.py)
            tx._data.master_addr = donor_address
            
            # Add optional message data
            tx._data.transfer.message_data = b'Staking Reward'
            
            # Get transaction data hash for signing (like create_transaction.py)
            tx_data = tx.get_data_hash()
            
            # Sign the transaction with donor wallet private key
            from qbitcoin.crypto.falcon import FalconSignature
            signature = FalconSignature.sign_message(tx_data, donor_private_key)
            tx._data.signature = signature
            
            # Update transaction hash after signing (like create_transaction.py)
            tx.update_txhash()
            
            # Submit transaction via gRPC to local node (like create_transaction.py)
            if self._submit_transaction_via_grpc(tx):
                logger.info("Reward transaction submitted via gRPC successfully: %s", bin2hstr(tx.txhash))
                return bin2hstr(tx.txhash)
            else:
                logger.error("Failed to submit reward transaction via gRPC")
                return None
            
        except Exception as e:
            logger.error("Error creating single reward transaction: %s", str(e))
            return None
            
    def _submit_transaction_via_grpc(self, tx) -> bool:
        """
        Submit transaction via gRPC to local node (like create_transaction.py does)
        
        Args:
            tx: The transaction to submit
            
        Returns:
            bool: True if successfully submitted, False otherwise
        """
        try:
            import grpc
            from qbitcoin.generated import qbit_pb2, qbit_pb2_grpc
            
            # Set up the gRPC connection to local node
            channel = grpc.insecure_channel("localhost:19009")
            stub = qbit_pb2_grpc.PublicAPIStub(channel)
            
            # Create the push transaction request (like create_transaction.py)
            push_transaction_req = qbit_pb2.PushTransactionReq(transaction_signed=tx.pbdata)
            
            logger.info("Submitting reward transaction via gRPC to local node...")
            push_transaction_resp = stub.PushTransaction(push_transaction_req, timeout=5)
            
            if push_transaction_resp.error_code == qbit_pb2.PushTransactionResp.SUBMITTED:
                logger.info("Reward transaction successfully submitted via gRPC!")
                return True
            else:
                logger.error("Reward transaction submission failed: %s", push_transaction_resp.error_description)
                return False
                
        except Exception as e:
            logger.error("Error submitting transaction via gRPC: %s", str(e))
            return False
            
        except Exception as e:
            logger.error("Error creating single reward transaction: %s", str(e))
            return None
        
    def _broadcast_staker_update(self, address: bytes, action: str, balance: int):
        """Broadcast staker updates to the network"""
        # Check if node is synced and has peer connections
        if not self.is_node_synced or not self.p2p_factory or self.p2p_factory.num_connections == 0:
            # Queue for later broadcast when node is synced
            with self.lock:
                self.pending_broadcasts.append((address, action, balance))
                logger.info("Queued staker %s: %s for broadcast when synced (connections: %d)", 
                           action, bin2hstr(address), 
                           self.p2p_factory.num_connections if self.p2p_factory else 0)
            return
            
        if self.p2p_factory:
            message = {
                'type': 'staker_update',
                'action': action,  # 'add' or 'remove'
                'address': bin2hstr(address),
                'balance': balance,
                'timestamp': time.time(),
                'node_id': 'local'  # Could be node identifier
            }
            try:
                # Send to all connected peers
                self.p2p_factory.broadcast_staker_message(message)
                logger.info("Broadcasted staker %s: %s", action, bin2hstr(address))
            except Exception as e:
                logger.error("Failed to broadcast staker update: %s", str(e))
        else:
            logger.debug("No P2P factory available for broadcasting staker update")
            
    def on_node_synced(self):
        """Called when the node becomes synced"""
        with self.lock:
            self.is_node_synced = True
            logger.info("Node synced - staking manager ready for P2P operations")
            
    def handle_peer_staker_message(self, message: dict, peer_ip: str):
        """Handle staker messages received from peers"""
        try:
            if message.get('type') != 'staker_update':
                return
                
            action = message.get('action')
            address_hex = message.get('address')
            
            if not address_hex:
                logger.warning("Invalid staker message from %s: missing address", peer_ip)
                return
                
            if not isinstance(address_hex, str):
                logger.error("Invalid address type from peer %s: expected str, got %s", peer_ip, type(address_hex))
                return
                
            try:
                address_bytes = hstr2bin(address_hex)
            except Exception as e:
                logger.error("Failed to convert address from peer %s: %s", peer_ip, str(e))
                return
            
            if action == 'add':
                # Verify balance from blockchain before adding
                balance = self.verify_staker_balance(address_bytes)
                if balance is None:
                    logger.warning("Invalid staker from peer %s: %s", peer_ip, address_hex)
                    return
                
                logger.info("Received valid staker add from peer %s: %s (verified balance: %d)", 
                           peer_ip, address_hex, balance)
                
                # Add the staker
                self.add_staker(address_bytes, peer_ip)
                
            elif action == 'remove':
                logger.info("Received staker remove from peer %s: %s", peer_ip, address_hex)
                self.remove_staker(address_bytes, f"removed by peer {peer_ip}")
            else:
                logger.warning("Unknown staker action from peer %s: %s", peer_ip, action)
                
        except Exception as e:
            logger.error("Error handling peer staker message from %s: %s", peer_ip, str(e))
            
    def get_stakers_for_sync(self) -> List[dict]:
        """Get all valid stakers for peer synchronization"""
        active_stakers = self.get_active_stakers()
        
        return [
            {
                'type': 'staker_update',
                'action': 'add',
                'address': bin2hstr(address),
                'balance': balance,
                'timestamp': time.time()
            }
            for address, balance in active_stakers.items()
        ]
            
     
    def get_validated_stakers_for_sync(self) -> List[dict]:
        """Get validated stakers for peer synchronization - same as get_stakers_for_sync"""
        return self.get_stakers_for_sync()
            
    def get_staking_statistics(self) -> dict:
        """Get current staking statistics"""
        active_stakers = self.get_active_stakers()
        total_stake = sum(active_stakers.values())
        
        return {
            'total_stakers': len(self.stakers),
            'active_stakers': len(active_stakers),
            'total_stake': total_stake,
            'donor_wallet': self.get_donor_wallet_address(),
            'donor_wallet_loaded': self.is_donor_wallet_loaded(),
            'last_reward_block': self.last_reward_block,
            'staking_enabled': self.dev_config.staking_enabled if self.dev_config else False,
            'reward_interval': self.dev_config.staking_reward_interval if self.dev_config else 60,
            'reward_amount_qbc': (self.dev_config.staking_reward_amount // 1000000000) if self.dev_config else 0,
            'min_balance_qbc': (self.dev_config.staking_min_balance // 1000000000) if self.dev_config else 0
        }
        
    def handle_new_block(self, block):
        """Handle new block events for reward distribution"""
        logger.debug("StakingManager.handle_new_block called for block %d", block.block_number)
        logger.debug("Last reward block: %d, Current block: %d, Interval: %d", 
                    self.last_reward_block, block.block_number, 
                    self.dev_config.staking_reward_interval if self.dev_config else 0)
        
        if self.should_distribute_rewards(block.block_number):
            logger.info("Attempting to distribute rewards for block %d", block.block_number)
            rewards = self.distribute_rewards(block.block_number)
            if rewards:
                logger.info("Distributed %d rewards for block %d", len(rewards), block.block_number)
            else:
                logger.warning("Failed to distribute rewards for block %d", block.block_number)
        else:
            logger.debug("Not time to distribute rewards for block %d", block.block_number)
            
    def request_staker_list_from_peer(self, peer_ip: str):
        """Request staker list from a specific peer (like requesting blocks)"""
        if self.p2p_factory:
            try:
                message = {
                    'type': 'staker_list_request',
                    'timestamp': time.time(),
                    'node_id': 'local'
                }
                self.p2p_factory.send_message_to_peer(peer_ip, message)
                logger.info("Requested staker list from peer: %s", peer_ip)
            except Exception as e:
                logger.error("Failed to request staker list from peer %s: %s", peer_ip, str(e))
                
    def handle_staker_list_request(self, peer_ip: str):
        """Handle staker list request from peer and send our staker list"""
        try:
            staker_list = self.get_stakers_for_sync()
            message = {
                'type': 'staker_list_response',
                'stakers': staker_list,
                'timestamp': time.time(),
                'node_id': 'local'
            }
            
            if self.p2p_factory:
                self.p2p_factory.send_message_to_peer(peer_ip, message)
                logger.info("Sent staker list (%d stakers) to peer: %s", len(staker_list), peer_ip)
                
        except Exception as e:
            logger.error("Failed to send staker list to peer %s: %s", peer_ip, str(e))
            
    def handle_staker_list_response(self, staker_list: List[dict], peer_ip: str):
        """Handle staker list response from peer - validate and add stakers"""
        if not staker_list:
            logger.info("Received empty staker list from peer: %s", peer_ip)
            return
            
        logger.info("Received staker list with %d stakers from peer: %s", len(staker_list), peer_ip)
        
        added_count = 0
        for staker_data in staker_list:
            try:
                address_hex = staker_data.get('address')
                if not address_hex:
                    continue
                    
                address_bytes = hstr2bin(address_hex)
                
                # Verify this staker is not already in our list (no duplicates)
                if address_bytes in self.stakers:
                    continue
                    
                # Verify balance from blockchain
                balance = self.verify_staker_balance(address_bytes)
                if balance is None:
                    logger.warning("Invalid staker from peer %s: %s (failed balance verification)", 
                                 peer_ip, address_hex)
                    continue
                    
                # Add the verified staker
                if self.add_staker(address_bytes, peer_ip):
                    added_count += 1
                    
            except Exception as e:
                logger.error("Error processing staker from peer %s: %s", peer_ip, str(e))
                
        logger.info("Added %d new verified stakers from peer: %s", added_count, peer_ip)
        
    def broadcast_staker_to_peers(self, address: bytes, action: str):
        """Broadcast staker add/remove to all connected peers"""
        if not self.p2p_factory or not self.is_node_synced:
            return
            
        try:
            message = {
                'type': 'staker_update',
                'action': action,  # 'add' or 'remove'
                'address': bin2hstr(address),
                'timestamp': time.time(),
                'node_id': 'local'
            }
            
            if self.p2p_factory.num_connections > 0:
                self.p2p_factory.broadcast_staker_message(message)
                logger.info("Broadcasted staker %s to peers: %s", action, bin2hstr(address))
            else:
                logger.debug("No peer connections for broadcasting staker %s: %s", action, bin2hstr(address))
                
        except Exception as e:
            logger.error("Failed to broadcast staker update: %s", str(e))
            
    def _monitor_rewards(self):
        """Monitor and distribute rewards automatically when interval is reached"""
        while self._reward_monitoring:
            try:
                if (self.dev_config and 
                    self.dev_config.staking_enabled and 
                    self.is_donor_wallet_loaded() and 
                    self.chain_manager):
                    
                    # Get current block number
                    try:
                        current_block = self.chain_manager.last_block
                        if current_block and self.should_distribute_rewards(current_block.block_number):
                            logger.info("Reward monitor: Time to distribute rewards for block %d", current_block.block_number)
                            rewards = self.distribute_rewards(current_block.block_number)
                            if rewards:
                                logger.info("Reward monitor: Successfully distributed %d rewards", len(rewards))
                            else:
                                logger.debug("Reward monitor: No rewards distributed (no active stakers or other issue)")
                    except Exception as e:
                        logger.error("Error in reward monitoring: %s", str(e))
                
                # Periodic staker validation (removes stakers with insufficient balance)
                try:
                    self.periodic_staker_validation()
                except Exception as e:
                    logger.error("Error in periodic staker validation: %s", str(e))
                
                # Check every 30 seconds
                time.sleep(30)
                
            except Exception as e:
                logger.error("Error in reward monitoring loop: %s", str(e))
                time.sleep(30)
                
    def stop(self):
        """Stop the staking manager"""
        self._reward_monitoring = False
        if hasattr(self, '_reward_thread') and self._reward_thread.is_alive():
            self._reward_thread.join(timeout=5)



from pyqrllib.pyqrllib import bin2hstr

from qbitcoin.core.State import State
from qbitcoin.core.StateContainer import StateContainer
from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
from qbitcoin.core.MultiSigAddressState import MultiSigAddressState
from qbitcoin.core.misc import logger
from qbitcoin.core.txs.Transaction import Transaction
from qbitcoin.core.FalconHelper import falcon_pk_to_address
from qbitcoin.crypto.misc import sha256


class MultiSigCreate(Transaction):
    """
    MultiSigCreate for the creation of multi sig wallet.
    """

    def __init__(self, protobuf_transaction=None):
        super(MultiSigCreate, self).__init__(protobuf_transaction)

    @property
    def signatories(self):
        return self._data.multi_sig_create.signatories

    @property
    def weights(self):
        return self._data.multi_sig_create.weights

    @property
    def total_weight(self):
        total_weight = 0
        for weight in self.weights:
            total_weight += weight
        return total_weight

    @property
    def threshold(self):
        return self._data.multi_sig_create.threshold

    def get_data_bytes(self):
        tmptxhash = (self.master_addr +
                     self.fee.to_bytes(8, byteorder='big', signed=False) +
                     self.threshold.to_bytes(8, byteorder='big', signed=False))

        for index in range(0, len(self.signatories)):
            tmptxhash = (tmptxhash +
                         self.signatories[index] +
                         self.weights[index].to_bytes(8, byteorder='big', signed=False))

        return tmptxhash

    def get_data_hash(self):
        return sha256(self.get_data_bytes())

    @staticmethod
    def create(signatories: list, weights: list, threshold: int, fee, xmss_pk, master_addr: bytes = None):
        multi_sig_create = MultiSigCreate()

        if master_addr:
            multi_sig_create._data.master_addr = master_addr

        multi_sig_create._data.public_key = bytes(xmss_pk)

        for signatory in signatories:
            multi_sig_create._data.multi_sig_create.signatories.append(signatory)

        for weight in weights:
            multi_sig_create._data.multi_sig_create.weights.append(weight)

        multi_sig_create._data.multi_sig_create.threshold = threshold
        multi_sig_create._data.fee = int(fee)

        multi_sig_create.validate_or_raise(verify_signature=False)

        return multi_sig_create

    def _validate_custom(self):
        if len(self.signatories) == 0:
            logger.warning("[MultiSigCreate] No Signatories found")
            return False

        for weight in self.weights:
            if weight == 0:
                logger.warning('Weight cannot be 0 - %s', weight)
                logger.warning('Invalid MultiSigCreate Transaction')
                return False

        if self.fee < 0:
            logger.warning('MultiSigCreate [%s] Invalid Fee = %d', bin2hstr(self.txhash), self.fee)
            return False

        if self.total_weight < self.threshold:
            logger.warning('[MultiSigCreate] Validation failed for %s because: Insufficient weight',
                           bin2hstr(self.txhash))
            logger.warning('Total weight: %s, Threshold: %s', self.total_weight, self.threshold)
            return False

        if len(set(self.signatories)) != len(self.signatories):
            logger.warning('[MultiSigCreate] Signatories list include duplicate signatories')
            return False

        if len(self.signatories) != len(self.weights):
            logger.warning('[MultiSigCreate] Mismatch number of signatories & weights')
            logger.warning('>> Length of signatories %s', len(self.signatories))
            logger.warning('>> Length of weights %s', len(self.weights))
            return False

        if not OptimizedAddressState.address_is_valid(self.addr_from):
            logger.warning('[MultiSigCreate] Invalid address addr_from: %s', bin2hstr(self.addr_from))
            return False

        for signatory in self.signatories:
            if not OptimizedAddressState.address_is_valid(signatory):
                logger.warning('[MultiSigCreate] Invalid address addr_to: %s', bin2hstr(signatory))
                return False

        return True

    # checks new tx validity based upon node statedb and node mempool.
    def _validate_extended(self, state_container: StateContainer):
        if state_container.block_number < state_container.current_dev_config.hard_fork_heights[0]:
            logger.warning("[MultiSigCreate] Hard Fork Feature not yet activated")
            return False

        if len(self.signatories) > state_container.current_dev_config.transaction_multi_output_limit:
            logger.warning('[MultiSigCreate] Number of signatories exceeds max limit')
            logger.warning('Number of Signatories %s', len(self.signatories))
            logger.warning('Number of Weights %s', len(self.weights))
            return False

        tx_balance = state_container.addresses_state[self.addr_from].balance

        if tx_balance < self.fee:
            logger.info('State validation failed for %s because: Insufficient funds', bin2hstr(self.txhash))
            logger.info('balance: %s, fee: %s, amount: %s', tx_balance, self.fee)
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)
        for signatory in self.signatories:
            addresses_set.add(signatory)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee, subtract=True)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)

        multi_sig_address_state = MultiSigAddressState.get_default(self.txhash,
                                                                   self.signatories,
                                                                   self.weights,
                                                                   self.threshold)
        state_container.addresses_state[multi_sig_address_state.address] = multi_sig_address_state
        state_container.paginated_tx_hash.insert(multi_sig_address_state, self.txhash)

        addr_from_pk = falcon_pk_to_address(self.PK)
        for index in range(0, len(self.signatories)):
            signatory = self.signatories[index]
            address_state = state_container.addresses_state[signatory]
            if signatory not in (self.addr_from, addr_from_pk):
                state_container.paginated_tx_hash.insert(address_state, self.txhash)
            state_container.paginated_multisig_address.insert(address_state, multi_sig_address_state.address)

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        multi_sig_address = MultiSigAddressState.generate_multi_sig_address(self.txhash)
        MultiSigAddressState.remove_multi_sig_address_state(state, multi_sig_address, state_container.batch)

        addr_from_pk = falcon_pk_to_address(self.PK)
        for index in range(0, len(self.signatories)):
            signatory = self.signatories[index]
            address_state = state_container.addresses_state[signatory]
            if signatory not in (self.addr_from, addr_from_pk):
                state_container.paginated_tx_hash.remove(address_state, self.txhash)
            state_container.paginated_multisig_address.remove(address_state, multi_sig_address)

        if multi_sig_address in state_container.addresses_state:
            del state_container.addresses_state[multi_sig_address]

        return self._revert_state_changes_for_PK(state_container)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import threading

from qbitcoin.core.Singleton import Singleton
from pyqrandomx import pyqrandomx


class QRandomX(object, metaclass=Singleton):

    def __init__(self):
        self.lock = threading.Lock()
        self._qrx = pyqrandomx.ThreadedQRandomX()

    def get_seed_height(self, block_number):
        return self._qrx.getSeedHeight(block_number)

    def hash(self, block_height, seed_height, seed_hash, blob):
        with self.lock:
            return bytes(self._qrx.hash(block_height, seed_height, seed_hash, blob, 0))
